from __future__ import annotations

import typing

import polars as pl

from ctc import spec
from ctc.toolbox import pl_utils


def interpolate(
    df: spec.DataFrame,
    *,
    index_column: str,
    start_index: int | None = None,
    end_index: int | None = None,
    pre_fill_values: typing.Mapping[str, typing.Any] | None = None,
) -> spec.DataFrame:
    """interpolate values in dataframe according to index column"""

    if len(df) == 0:
        return df

    index = df[index_column]

    # remove redundant rows wrt index_column
    df = df.filter(pl.col(index_column).cumcount().over(index_column) == 0)

    # validate index
    if index.dtype not in [pl.Int8, pl.Int16, pl.Int32, pl.Int64]:
        raise Exception('index must be an integer column')
    if not index.series_equal(index.sort()):
        raise Exception('index must be a sorted column')
    if start_index is not None:
        if start_index > index[0]:
            raise Exception('start_index is larger than first index in df')
    else:
        start_index = index[0]
    if end_index is not None:
        if end_index < index[-1]:
            raise Exception('end_index is smaller than last index in df')
    else:
        end_index = index[-1]

    # build new index and initial values
    if start_index < index[0]:
        if pre_fill_values is None:
            raise Exception(
                'must specify pre_fill_values if start_index < index[0]'
            )
        initial_values: typing.MutableMapping[
            str, None | typing.Sequence[typing.Any]
        ] = {k: [v] for k, v in pre_fill_values.items()}
        initial_values[index_column] = [start_index]
        for column in df.columns:
            if column not in initial_values and column != index_column:
                if df[column].dtype == pl.Object:
                    initial_values[column] = [None for i in range(len(df))]
                else:
                    initial_values[column] = None
        initial_df = pl.DataFrame(initial_values, schema=df.schema)
        new_index = range(start_index + 1, end_index + 1)
    else:
        initial_df = None
        new_index = range(start_index, end_index + 1)

    # create interpolation rows
    new_data: typing.MutableMapping[str, typing.Any] = {
        column: None for column in df.columns if column != index_column
    }
    new_data[index_column] = pl.DataFrame(
        pl.Series(index_column, new_index)
    ).filter(~pl.col(index_column).is_in(df[index_column]))[index_column]
    for column in df.columns:
        if (pre_fill_values is None or column not in pre_fill_values) and df[
            column
        ].dtype == pl.Object:
            new_data[column] = [
                None for i in range(len(new_data[index_column]))
            ]
    new_df = pl.DataFrame(new_data, schema=df.schema)

    # concat, sort, and fill
    if initial_df is not None:
        dfs = [initial_df, df, new_df]
    else:
        dfs = [df, new_df]
    concated = pl_utils.concat(dfs).sort(index_column)

    # handle pl.Object columns separately
    filled = {}
    for column in concated.columns:
        if concated[column].dtype == pl.Object:
            new = []
            current = None
            for item in concated[column].to_list():
                if item is not None:
                    current = item
                new.append(current)
            filled[column] = pl.Series(new, dtype=pl.Object)
        else:
            filled[column] = concated[column].fill_null(strategy='forward')

    return pl.DataFrame(filled)



import pytest
import toolcli

from ctc.cli import cli_run


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommands_verbose_can_use_single_v(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    for arg in command_spec.get('args', []):
        if arg.get('name') == '--verbose':
            raise Exception(
                'in '
                + str(spec_reference)
                + ' should use both --verbose AND -v for cli args'
            )


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommand_options_use_dashes_not_underscores(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    for arg in command_spec.get('args', []):
        name = arg['name']
        if isinstance(name, str):
            name_list = [name]
        else:
            name_list = name

        for subname in name_list:
            if subname.startswith('--') and '_' in subname:
                raise Exception(
                    'error in '
                    + spec_reference
                    + ' cli args should use dashes not underscores'
                )


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommands_use_export_instead_of_output(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    problem_commands = []
    for arg in command_spec.get('args', []):
        if arg.get('name') == '--output':
            raise Exception(
                'these commands should use the name --export instead of --output: '
                + str(problem_commands)
            )



from __future__ import annotations

import toolcli
import toolstr

from ctc import cli
from ctc import config
from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_chains_command,
        'help': 'display information about configured chains',
        'args': [
            {
                'name': ['--verbose', '-v'],
                'help': 'display additional information including RPC urls',
                'action': 'store_true',
            },
        ],
        'examples': ['', '--verbose'],
    }


async def async_chains_command(verbose: bool) -> None:

    default_network = config.get_default_network()
    all_networks = list(evm.get_networks().values())
    all_networks = sorted(all_networks, key=lambda network: network['chain_id'])

    # providers
    providers = config.get_providers()
    providers_by_chain = {
        provider['network']: provider for provider in providers.values()
    }

    already_used = set()
    labels = [
        'name',
        'chain_id',
        'block explorer',
        'default',
        'has rpc',
    ]
    if verbose:
        labels[4] = 'rpc'

    rows = []
    for network in all_networks:

        if network['chain_id'] in already_used:
            continue
        already_used.add(network['chain_id'])

        is_default = '✓' if default_network == network['chain_id'] else ''

        if not verbose:
            has_provider = '✓' if network['chain_id'] in providers_by_chain else ''
        else:
            if network['chain_id'] in providers_by_chain:
                has_provider = providers_by_chain[network['chain_id']]['url']
            else:
                has_provider = ''

        row = [
            network['name'],
            network['chain_id'],
            network['block_explorer'],
            is_default,
            has_provider,
        ]
        rows.append(row)

    rows = sorted(rows, key=lambda row: (row[3], row[4], -row[1]), reverse=True)  # type: ignore

    styles = cli.get_cli_styles()

    toolstr.print_text_box('EVM Chains', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=labels,
        column_justify={
            'default': 'center',
            'has rpc': 'center',
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'chain_id': styles['description'],
            'block explorer': styles['metavar'],
            'default': styles['title'],
            'has rpc': styles['title'],
        },
    )



from __future__ import annotations

import json
import os
import shutil
import tempfile

import toolcli
import toolstr

from ctc import cli
import ctc.config


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': edit_config_command,
        'help': 'edit config values',
        'hidden': True,
        'examples': [''],
    }


def edit_config_command() -> None:
    import subprocess

    editor = os.environ.get('EDITOR')
    if editor is None or editor == '':
        editor = 'editor'
        print('$EDITOR env var not set, attempting to use default editor')

    config_path = ctc.config.get_config_path()

    styles = cli.get_cli_styles()
    toolstr.print('editing config in editor', style=styles['description'])
    cli.print_bullet(key='editor', value=editor)
    cli.print_bullet(key='config_path', value=config_path)
    print()

    tempdir = tempfile.mkdtemp()
    temppath = os.path.join(tempdir, 'config.json')
    with open(config_path, 'r') as f:
        config_data = json.load(f)
    with open(temppath, 'w') as f:
        json.dump(config_data, f, indent=4, sort_keys=True)
    subprocess.call([editor, temppath])
    shutil.copy2(temppath, config_path)
    print('done editing config')



# TODO: update with additional proxy types, see:
# - https://github.com/ApeWorX/ape/pull/750
# - https://docs.openzeppelin.com/contracts/4.x/api/proxy

from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict, Literal

    class ProxyAddressMetadata(TypedDict):
        implementation: spec.Address | None
        proxy_type: Literal[
            'eip897',
            'eip1967-logic',
            'eip1967-beacon',
            'openzeppelin',
            'gnosis_safe',
        ] | None


async def async_get_proxy_implementation(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address | None:
    """return implementation address of proxy contract"""

    proxy_metadata = await async_get_proxy_metadata(
        contract_address=contract_address, context=context, block=block
    )

    return proxy_metadata['implementation']


async def async_get_proxy_metadata(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> ProxyAddressMetadata:
    """return metadata of proxy address"""

    # try eip897
    try:
        eip897_address = await _async_get_eip897_implementation(
            contract_address=contract_address,
            context=context,
            block=block,
        )
        if eip897_address is not None:
            return {
                'implementation': eip897_address,
                'proxy_type': 'eip897',
            }
    except (spec.exceptions.rpc_exceptions.RpcException, Exception):
        pass

    # try eip1967 logic
    eip1967_logic_address = await _async_get_eip1967_proxy_logic_address(
        contract_address=contract_address,
        context=context,
        block=block,
    )
    if eip1967_logic_address != '0x0000000000000000000000000000000000000000':
        return {
            'implementation': eip1967_logic_address,
            'proxy_type': 'eip1967-logic',
        }

    # try eip1967 beacon
    eip1967_beacon_address = await _async_get_eip1967_proxy_beacon_address(
        contract_address=contract_address,
        context=context,
        block=block,
    )
    if eip1967_beacon_address is not None:
        return {
            'implementation': eip1967_beacon_address,
            'proxy_type': 'eip1967-beacon',
        }

    # try openzeppelin
    openzeppelin_address = await _async_get_oz_proxy_address(
        contract_address=contract_address,
        context=context,
        block=block,
    )
    if openzeppelin_address is not None:
        return {
            'implementation': openzeppelin_address,
            'proxy_type': 'openzeppelin',
        }

    # try gnosis proxy
    gnosis_proxy_address = await _async_get_gnosis_safe_proxy_address(
        contract_address,
        context=context,
        block=block,
    )
    if gnosis_proxy_address is not None:
        return {
            'implementation': gnosis_proxy_address,
            'proxy_type': 'gnosis_safe',
        }

    return {
        'implementation': None,
        'proxy_type': None,
    }


#
# # eip897
#


async def _async_get_eip897_proxy_type(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int | None:

    from ctc import rpc

    function_abi: spec.FunctionABI = {
        'name': 'proxyType',
        'type': 'function',
        'inputs': [],
        'outputs': [
            {'name': 'proxyTypeId', 'type': 'uint256'},
        ],
    }

    result = await rpc.async_eth_call(
        to_address=contract_address,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if result is not None and not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def _async_get_eip897_implementation(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address:

    from ctc import rpc

    function_abi: spec.FunctionABI = {
        'name': 'implementation',
        'type': 'function',
        'inputs': [],
        'outputs': [
            {'name': 'codeAddr', 'type': 'address'},
        ],
    }

    result = await rpc.async_eth_call(
        to_address=contract_address,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


#
# # eip1967
#


async def _async_get_eip1967_proxy_logic_address(
    contract_address: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> spec.Address:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.implementation')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    from ctc import rpc

    position = (
        '0x360894a13ba1a3210667c828492db98dca3e2076cc3735a920a3ca505d382bbc'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    return '0x' + result[-40:]


async def _async_get_eip1967_proxy_beacon_address(
    contract_address: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> spec.Address | None:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.beacon')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    from ctc import rpc

    position = (
        '0xa3f0ad74e5423aebfd80d3ef4346578335a9a72aeaee59ff6cb3582b35133d50'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    beacon = '0x' + result[-40:]
    if beacon == '0x0000000000000000000000000000000000000000':
        return None

    implementation_abi: spec.FunctionABI = {
        'name': 'implementation',
        'type': 'function',
        'inputs': [],
        'outputs': [{'type': 'address'}],
    }

    try:
        proxy_in_beacon = await rpc.async_eth_call(
            to_address=beacon,
            function_abi=implementation_abi,
            context=context,
        )
    except Exception:
        proxy_in_beacon = None

    if not isinstance(proxy_in_beacon, str) and proxy_in_beacon is not None:
        raise Exception('bad rpc data')
    return proxy_in_beacon


async def _async_get_eip1967_proxy_admin_address(
    contract_address: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> spec.Address:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.admin')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    from ctc import rpc

    position = (
        '0xa3f0ad74e5423aebfd80d3ef4346578335a9a72aeaee59ff6cb3582b35133d50'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    return '0x' + result[-40:]


# async def async_get_eip1967_history() -> None:
#     raise NotImplementedError(
#         'use events here, see https://docs.openzeppelin.com/contracts/4.x/api/proxy#BeaconProxy'
#     )


#
# # openzeppelin os
#


async def _async_get_oz_proxy_address(
    contract_address: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> spec.Address | None:

    from ctc import rpc

    position = (
        '0x7050c9e0f4ca769c69bd3a8ef740bc37934f8e2c036e5a723fd8ee048ed3f8c3'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    address = '0x' + result[-40:]

    if address == '0x0000000000000000000000000000000000000000':
        return None
    else:
        return address


#
# # gnosis
#


async def _async_get_gnosis_safe_proxy_address(
    contract_address: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    confirm_bytecode: bool = True,
    context: spec.Context = None,
) -> spec.Address | None:

    from ctc import rpc

    if confirm_bytecode:
        gnosis_proxy_code = '0x608060405273ffffffffffffffffffffffffffffffffffffffff600054167fa619486e0000000000000000000000000000000000000000000000000000000060003514156050578060005260206000f35b3660008037600080366000845af43d6000803e60008114156070573d6000fd5b3d6000f3fea2646970667358221220d1429297349653a4918076d650332de1a1068c5f3e07c5c82360c277770b955264736f6c63430007060033'
        bytecode = await rpc.async_eth_get_code(
            contract_address, block_number=block, context=context
        )
        if bytecode != gnosis_proxy_code:
            return None

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position='0x0',
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    return '0x' + result[-40:]




import os
import configparser


def test_tox_legacy_config():
    import toml

    repo_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    path = os.path.join(repo_root, 'pyproject.toml')
    print(path)
    with open(path) as f:
        config = toml.load(f)

    # load current deps
    current_deps = {}
    for dep_str in config['project']['dependencies']:
        if '>=' not in dep_str:
            raise Exception('need min version for dependency: ' + str(dep_str))
        dep_name, version_tail = dep_str.split(' >=')
        min_version, max_version = version_tail.split(', <')
        current_deps[dep_name] = tuple(
            number for number in min_version.split('.')
        )

    # load legacy deps
    legacy_parser = configparser.ConfigParser()
    legacy_parser.read_string(config['tool']['tox']['legacy_tox_ini'])
    legacy_deps_str = legacy_parser.get("testenv:py37-legacy", "deps")
    legacy_deps_strs = legacy_deps_str.split('\n')
    legacy_deps = {}
    for legacy_dep_str in legacy_deps_strs:
        if legacy_dep_str == '':
            continue

        if '==' not in legacy_dep_str:
            continue

        dep_name, version = legacy_dep_str.split('==')
        legacy_deps[dep_name] = tuple(
            number for number in version.split('.')
        )

    # assert that legacy uses the min version number of each dependency
    for dep_name in legacy_deps.keys():
        if dep_name not in current_deps:
            raise Exception(
                dep_name
                + ' in legacy dependencies but not in current dependencies'
            )
        assert current_deps[dep_name] == legacy_deps[dep_name], (
            dep_name
            + ' has different versions in legacy vs current test environments'
        )
    for dep_name in current_deps:
        if dep_name not in legacy_deps:
            raise Exception(
                dep_name
                + ' in current dependencies but not in legacy dependencies'
            )



from __future__ import annotations

import functools
import os
import typing

from ctc import config
from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


def resolve_provider(
    provider: spec.ProviderReference,
    other_providers: typing.Mapping[str, spec.Provider] | None = None,
) -> spec.Provider:
    """try to lookup existing provider, and if it does not exist, create one"""
    if isinstance(provider, str):
        if other_providers is not None:
            candidates = list(other_providers.values())
        else:
            candidates = list(config.get_providers().values())
        try:
            # try provider name
            return find_provider(name=provider, candidates=candidates)
        except LookupError:
            try:
                # try provider url
                return find_provider(url=provider, candidates=candidates)
            except LookupError:
                # create new provider
                return create_provider(
                    url=provider, other_providers=other_providers
                )
    elif isinstance(provider, dict):
        return create_provider(other_providers=other_providers, **provider)
    else:
        raise Exception('unknown provider format: ' + str(type(provider)))


def create_provider(
    *,
    url: str | None = None,
    network: spec.NetworkReference | None = None,
    name: str | None = None,
    protocol: Literal['http', 'wss', 'ipc'] | None = None,
    session_kwargs: typing.Mapping[str, typing.Any] | None = None,
    chunk_size: int | None = None,
    convert_reverts_to_none: bool = True,
    validate_chain_id: bool = True,
    disable_batch_requests: bool = False,
    other_providers: typing.Mapping[str, spec.Provider] | None = None,
) -> spec.Provider:
    """create provider"""

    # url
    if url is None:
        if network is None:
            network = config.get_default_network()
        default_provider = config.get_network_default_provider(network)
        if default_provider is None:
            raise Exception('could not determine url for provider')
        url = default_provider['url']
    if not url.startswith('http'):
        if '://' in url:
            raise Exception('only http and https currently supported')

        # try http
        try:
            http_url = 'http://' + url
            _sync_get_chain_id(http_url)
            url = http_url
        except Exception:

            # try https
            try:
                https_url = 'https://' + url
                _sync_get_chain_id(https_url)
                url = https_url
            except Exception:
                raise Exception(
                    'provider url does not specify http or https and provider not responding to requests. make sure the url includes any relevant port numbers and http/https prefixes'
                )

    # network
    if network is None or validate_chain_id:
        actual_chain_id = _sync_get_chain_id(url)
    if network is None:
        network = actual_chain_id
    else:
        network = evm.get_network_chain_id(network)
        if validate_chain_id and network != actual_chain_id:
            raise Exception('provider network does not match given network')

    # name
    if other_providers is None:
        other_providers = config.get_providers()
    if name is not None:
        if name in other_providers:
            raise Exception(
                'provider with this name already exists, simply use provider=<provider_name> in arg or context'
            )
    if name is None:
        try:
            name = url.split('://')[1].split('/')[0] + '__' + str(network)
        except Exception:
            raise Exception(
                'could not determine name for provider, make sure url is formatted properly'
            )
        if name in other_providers:
            raise Exception(
                'could not determine unique name for new provider, specify a name argument'
            )

    # protocol
    if protocol is None:
        if url.startswith('http'):
            protocol = 'http'
        elif url.startswith('wss'):
            protocol = 'wss'
        elif url.startswith('ipc'):
            protocol = 'ipc'
        else:
            raise Exception('unknown protocol used by provider')

    if session_kwargs is None:
        session_kwargs = {}

    provider: spec.Provider = {
        'name': name,
        'network': network,
        'protocol': protocol,
        'url': url,
        'session_kwargs': session_kwargs,
        'chunk_size': chunk_size,
        'convert_reverts_to_none': convert_reverts_to_none,
        'disable_batch_requests': disable_batch_requests,
    }

    return provider


def find_provider(
    *,
    name: str | None = None,
    network: str | int | None = None,
    url: str | None = None,
    candidates: typing.Sequence[spec.Provider] | None = None,
) -> spec.Provider:
    """find provider that matches inputs from list of candidates

    if no candidates provided, use providers in config
    """

    from ctc import config
    from ctc.toolbox import search_utils

    if candidates is None:
        candidates = list(config.get_providers().values())

    # build query
    query: typing.MutableMapping[str, str | int] = {}
    if name is None and network is None and url is None:
        raise Exception('specify network name or network or url')
    if name is not None:
        query['name'] = name
    if network is not None:
        if isinstance(network, str):
            from ctc import evm

            network = evm.get_network_chain_id(network)
        query['network'] = network
    if url is not None:
        query['url'] = url

    entries = search_utils.get_matching_entries(
        sequence=candidates, query=query
    )
    if len(entries) == 1:
        return entries[0]
    elif len(entries) > 1:
        if network is not None:
            provider = config.get_network_default_provider(network=network)
            if provider is not None:
                return provider
        raise LookupError('too many candidate providers found')
    else:
        raise LookupError('could not detect suitable RPC provider')


# def get_provider(provider: spec.ProviderReference = None) -> spec.Provider:
#     import ctc.config

#     if provider is None:

#         # case: return default provider
#         return ctc.config.get_default_provider()

#     elif isinstance(provider, str):

#         # case: provider specified as url
#         if provider.startswith('http'):
#             if ctc.config.has_provider(url=provider):
#                 return ctc.config.get_provider(url=provider)
#             else:
#                 try:
#                     network = _sync_get_chain_id(provider)
#                 except Exception:
#                     raise Exception('could not determine chain_id of provider')

#                 return {
#                     'name': None,
#                     'network': network,
#                     'protocol': 'http',
#                     'url': provider,
#                     'session_kwargs': {},
#                     'chunk_size': None,
#                     'convert_reverts_to_none': False,
#                 }

#         # case: provider specified as name in config
#         elif ctc.config.has_provider(name=provider):
#             return ctc.config.get_provider(name=provider)

#         else:
#             raise Exception('unknown provider format: ' + str(provider))

#     elif isinstance(provider, dict):

#         # case: partial provider
#         if set(spec.provider_keys).issubset(set(provider.keys())):
#             return provider  # type: ignore

#         else:
#             import copy

#             selection_keys = ['name', 'network', 'protocol']
#             if any(provider.get(key) is not None for key in selection_keys):
#                 base_provider = ctc.config.get_provider(
#                     name=provider.get('name'),
#                     network=provider.get('network'),
#                     protocol=provider.get('protocol'),
#                 )
#             else:
#                 base_provider = ctc.config.get_default_provider()

#             data = {k: v for k, v in provider.items() if v is not None}
#             if typing.TYPE_CHECKING:
#                 non_none_keys = typing.cast(spec.Provider, data)
#             else:
#                 non_none_keys = data
#             full_provider: spec.Provider = copy.copy(base_provider)
#             full_provider.update(non_none_keys)  # type: ignore

#             return full_provider

#     else:
#         raise Exception('unknown provider type: ' + str(type(provider)))


def _get_provider_id(provider: spec.Provider) -> spec.ProviderId:
    """return a unique identifier for provider within this process"""

    session_kwargs = provider.get('session_kwargs')
    if session_kwargs is None:
        session_kwargs = {}

    return (os.getpid(), provider['url'], tuple(session_kwargs.items()))


@functools.lru_cache()
def _sync_get_chain_id(provider_url: str) -> int:
    """synchronously obtain chain_id of provider"""
    import json
    import urllib.request

    data = {'jsonrpc': '2.0', 'method': 'eth_chainId', 'params': [], 'id': 1}
    encoded_data = json.dumps(data).encode()
    request = urllib.request.Request(
        provider_url,
        data=encoded_data,
        headers={
            'User-Agent': 'python3',
            'Content-Type': 'application/json',
        },
    )
    response = urllib.request.urlopen(request, timeout=5)
    response_data = json.loads(response.read().decode())
    raw_chain_id = response_data['result']
    return evm.binary_convert(raw_chain_id, 'integer')




from __future__ import annotations

import typing

import polars as pl

from ctc import spec


def set_column_display_width(width: int = 70) -> None:
    pl.Config.set_fmt_str_lengths(width)


def get_glob_total_bytes(path_template: str) -> int:
    import glob
    import os

    total_size = 0
    for path in glob.glob(path_template):
        total_size += os.path.getsize(path)
    return total_size


def get_series_distribution(series: spec.Series) -> spec.DataFrame:
    dist = (
        series.value_counts()
        .sort('counts', descending=True)
        .rename({'counts': 'count'})
    )
    dist = dist.with_columns(
        pl.col('count').cumsum().alias('cumcount'),
        (pl.col('count') / series.shape[0]).alias('pdf'),
    )
    dist = dist.with_columns(pl.col('pdf').cumsum().alias('cdf'))
    dist.insert_at_idx(1, dist.with_row_count()['row_nr'].alias('rank') + 1)
    return dist


def create_series_summary(series: pl.Series) -> typing.Mapping[str, typing.Any]:
    import toolstr

    if series.dtype == pl.datatypes.Object:
        n_unique = len(set(series.to_list()))
    elif isinstance(series.dtype, pl.List):
        n_unique = len(set(tuple(item) for item in series.to_list()))
    else:
        n_unique = series.n_unique()
    n_rows = len(series)
    n_null = series.is_null().sum()

    # bytes
    n_bytes: int | None = 0
    bytes_per_row = 0
    if series.dtype == pl.datatypes.Int8:
        bytes_per_row = 1
    elif series.dtype == pl.datatypes.Int16:
        bytes_per_row = 2
    elif series.dtype == pl.datatypes.Int32:
        bytes_per_row = 4
    elif series.dtype == pl.datatypes.Int64:
        bytes_per_row = 8
    elif series.dtype == pl.datatypes.Float32:
        bytes_per_row = 4
    elif series.dtype == pl.datatypes.Float64:
        bytes_per_row = 8
    elif series.dtype == pl.datatypes.Utf8:
        bytes_per_row = 8
        str_lengths = series.str.lengths()
        str_lengths_sum = typing.cast(int, str_lengths.sum())
        if str_lengths_sum is not None and n_bytes is not None:
            n_bytes += str_lengths_sum
    elif series.dtype == pl.datatypes.Categorical:
        bytes_per_row = 4
        str_lengths = series.unique().str.lengths()
        str_lengths_sum = typing.cast(int, str_lengths.sum())
        if str_lengths_sum is not None and n_bytes is not None:
            n_bytes += str_lengths_sum
    else:
        n_bytes = None
    if n_bytes is not None:
        n_bytes += bytes_per_row * n_rows

    # human-readable bytes
    if n_bytes is not None:
        h_bytes = toolstr.format_nbytes(n_bytes)
    else:
        h_bytes = None

    # unique compression
    if n_unique < 2**16:
        unique_key_bytes = 2
    elif n_unique < 2**32:
        unique_key_bytes = 4
    else:
        unique_key_bytes = 8
    if series.dtype == pl.datatypes.Utf8:
        unique_lengths = series.unique().str.lengths()
        unique_compress_bytes = (
            unique_key_bytes * n_rows + (unique_key_bytes + 8) * n_unique
        )
        unique_lengths_sum = typing.cast(int, unique_lengths.sum())
        if unique_lengths_sum is not None:
            unique_compress_bytes += unique_lengths_sum
    else:
        unique_compress_bytes = (
            unique_key_bytes + bytes_per_row
        ) * n_unique + unique_key_bytes * n_rows
    if n_bytes is not None and n_bytes != 0:
        unique_compression_factor = unique_compress_bytes / n_bytes
    else:
        unique_compression_factor = None

    # binary compress size
    binary_n_bytes = None
    binary_h_bytes = None
    binary_unique_n_bytes = None
    binary_unique_h_bytes = None
    binary_unique_compression = None
    if series.dtype == pl.datatypes.Utf8:
        try:
            is_hex = True
        except Exception:
            is_hex = False
        if is_hex:
            binary_lengths = (str_lengths - 2) / 2
            binary_n_bytes = 8 * n_rows
            binary_lengths_sum = typing.cast(int, binary_lengths.ceil().sum())
            if binary_lengths_sum is not None:
                binary_n_bytes += binary_lengths_sum
            binary_h_bytes = toolstr.format_nbytes(binary_n_bytes)

            binary_unique_n_bytes = (
                unique_key_bytes * n_rows + (unique_key_bytes + 8) * n_unique
            )
            unique_lengths_sum = typing.cast(
                int, ((unique_lengths - 2) / 2).ceil().sum()
            )
            if unique_lengths_sum is not None:
                binary_unique_n_bytes += unique_lengths_sum

            binary_unique_h_bytes = toolstr.format_nbytes(binary_unique_n_bytes)

            if n_bytes is not None and n_bytes != 0:
                binary_unique_compression = binary_unique_n_bytes / n_bytes
            else:
                binary_unique_compression = None

    return {
        "column": series.name,
        "n_unique": n_unique,
        "%_unique": n_unique / n_rows if n_rows > 0 else None,
        "dtype": str(series.dtype),
        "n_null": n_null,
        "%_null": n_null / n_rows if n_rows > 0 else None,
        "n_bytes": n_bytes,
        "h_bytes": h_bytes,
        'binary_n_bytes': binary_n_bytes,
        'binary_h_bytes': binary_h_bytes,
        "key_size": unique_key_bytes,
        'unique_n_bytes': unique_compress_bytes,
        "unique_h_bytes": toolstr.format_nbytes(unique_compress_bytes),
        'unique_compression': unique_compression_factor,
        'binary_unique_n_bytes': binary_unique_n_bytes,
        'binary_unique_h_bytes': binary_unique_h_bytes,
        'binary_unique_compression': binary_unique_compression,
    }


def create_dataframe_summary(df: spec.DataFrame) -> spec.DataFrame:
    summary = [create_series_summary(df[column]) for column in df.columns]
    return pl.DataFrame(summary)


def print_dataframe_summary(
    df: spec.DataFrame, title: str | None = None
) -> None:
    import toolstr

    summary = create_dataframe_summary(df)

    if title is None:
        title = 'Dataframe Summary'

    toolstr.print_text_box(title)
    non_object_df = df.select(
        column
        for column in df.columns
        if column not in df.select(pl.col(pl.Object))
    )
    rows: typing.Sequence[typing.Sequence[typing.Any]] = [
        ['n_rows', len(df)],
        ['n_columns', len(df.columns)],
        [
            'estimate_bytes()',
            toolstr.format_nbytes(non_object_df.estimated_size()),
        ],
    ]
    toolstr.print_table(rows, indent=4)
    print()

    labels = summary.columns
    column_formats: typing.Mapping[str, typing.Mapping[str, typing.Any]] = {
        "%_unique": {"percentage": True, "decimals": 1},
        "%_null": {"percentage": True, "decimals": 1},
        "unique_compression": {"percentage": True, "decimals": 1},
        'binary_unique_compression': {"percentage": True, "decimals": 1},
    }
    toolstr.print_header('Column summary')
    toolstr.print_dataframe_as_table(
        summary, indent=4, column_formats=column_formats
    )

    print()
    toolstr.print_header('Byte summary')
    n_byte_columns = [
        column for column in df.columns if column.endswith('n_bytes')
    ]
    byte_totals = summary[n_byte_columns].sum()
    rows = [
        [
            name,
            toolstr.format_nbytes(byte_totals[name][0]),
            byte_totals[name][0] / summary["n_bytes"].sum(),
        ]
        for name in byte_totals.columns
    ]
    labels = ["name", "n_bytes", "%"]
    column_formats = {'%': {'percentage': True, 'decimals': 1}}
    toolstr.print_table(
        rows,
        labels=labels,
        column_formats=column_formats,  # type: ignore
        indent=4,
    )




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import toolsql


fourbyte_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'function_signatures': {
            'columns': [
                {
                    'name': 'id',
                    'type': 'Integer',
                    'primary': True,
                },
                {
                    'name': 'created_at',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'hex_signature',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'text_signature',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'bytes_signature',
                    'type': 'Binary',
                    'index': True,
                },
            ],
        },
        'event_signatures': {
            'columns': [
                {
                    'name': 'id',
                    'type': 'Integer',
                    'primary': True,
                },
                {
                    'name': 'created_at',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'hex_signature',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'text_signature',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'bytes_signature',
                    'type': 'Binary',
                    'index': True,
                },
            ],
        },
    },
}



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import toolsql

from ctc import rpc
from ctc import spec

from . import management
from . import schemas


if typing.TYPE_CHECKING:
    T = typing.TypeVar('T', int, spec.RPCBlock, spec.DBBlock)


async def async_is_block_fully_confirmed(
    block: int,
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection | None,
) -> bool:
    """must pass conn=None if not planning on using db"""

    # check whether block is older than newest block in db
    if conn is not None:
        max_db_block = await _async_get_max_block_number_in_db(
            conn=conn, context=context
        )
        if max_db_block is not None and block < max_db_block:
            return True

    # check whether block has enough confirmations
    rpc_latest_block = await rpc.async_eth_block_number(context=context)
    required_confirmations = management.get_required_confirmations(
        context=context
    )
    return bool(block <= rpc_latest_block - required_confirmations)


async def async_filter_fully_confirmed_blocks(
    blocks: typing.Sequence[T],
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection | None,
    latest_block_number: int | None = None,
) -> typing.Sequence[T]:

    if len(blocks) == 0:
        return []

    block_numbers = []
    for block in blocks:
        if isinstance(block, dict):
            block_numbers.append(block['number'])
        elif isinstance(block, int):
            block_numbers.append(block)
        else:
            raise Exception('unknown block format')

    # check whether all blocks older than newest block in db
    max_block_number = max(block_numbers)
    if latest_block_number is not None:
        max_db_block: int | None = latest_block_number
    else:
        if conn is None:
            raise Exception('must provide conn or latest block number')
        max_db_block = await _async_get_max_block_number_in_db(
            context=context, conn=conn
        )
    if max_db_block is not None and max_db_block > max_block_number:
        return blocks

    # check whether blocks have enough confirmations
    if latest_block_number is not None:
        rpc_latest_block = latest_block_number
    else:
        rpc_latest_block = await rpc.async_eth_block_number(context=context)
    required_confirmations = management.get_required_confirmations(
        context=context,
    )
    max_allowed_block = rpc_latest_block - required_confirmations
    return [
        block
        for block, block_number in zip(blocks, block_numbers)
        if block_number <= max_allowed_block
    ]


async def _async_get_max_block_number_in_db(
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> int | None:
    return await schemas.async_select_max_block_number(
        conn=conn, context=context
    )




from __future__ import annotations

import typing

from ctc import evm

if typing.TYPE_CHECKING:
    import toolcli


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_abi_diff_command,
        'help': 'display diff of two contract ABI\'s',
        'args': [
            {
                'name': 'first_address',
                'help': 'first contract address',
            },
            {
                'name': 'second_address',
                'help': 'second contract address',
            },
            {
                'name': '--functions',
                'help': 'show functions only',
                'action': 'store_true',
            },
            {
                'name': '--events',
                'help': 'show events only',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419 0xf79d6afbb6da890132f9d7c355e3015f15f3406f',
        ],
    }


async def async_abi_diff_command(
    *,
    first_address: str,
    second_address: str,
    functions: bool,
    events: bool,
) -> None:
    first_abi = await evm.async_get_contract_abi(first_address)
    second_abi = await evm.async_get_contract_abi(second_address)
    evm.print_contract_abi_diff(
        first_contract_abi=first_abi,
        second_contract_abi=second_abi,
        first_name=first_address,
        second_name=second_address,
        functions_only=functions,
        events_only=events,
    )



from __future__ import annotations

import typing

from ctc import spec


@typing.overload
async def async_get_eth_balance(
    address: spec.Address,
    *,
    normalize: typing.Literal[False],
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> int:
    ...


@typing.overload
async def async_get_eth_balance(
    address: spec.Address,
    *,
    normalize: typing.Literal[True] = True,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> float:
    ...


@typing.overload
async def async_get_eth_balance(
    address: spec.Address,
    *,
    normalize: bool,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> int | float:
    ...


async def async_get_eth_balance(
    address: spec.Address,
    *,
    normalize: bool = True,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> typing.Union[int, float]:
    """get ETH balance"""

    from ctc import rpc

    result = await rpc.async_eth_get_balance(
        address=address,
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    balance: int | float = result

    if normalize:
        balance /= 1e18

    return balance


@typing.overload
async def async_get_eth_balance_by_block(
    address: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[False],
    context: spec.Context = None,
) -> list[int]:
    ...


@typing.overload
async def async_get_eth_balance_by_block(
    address: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[True] = True,
    context: spec.Context = None,
) -> list[float]:
    ...


@typing.overload
async def async_get_eth_balance_by_block(
    address: spec.Address,
    *,
    normalize: bool = True,
    blocks: typing.Sequence[spec.BlockNumberReference],
    context: spec.Context = None,
) -> typing.Union[list[int], list[float]]:
    ...


async def async_get_eth_balance_by_block(
    address: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[list[int], list[float]]:
    """get historical ETH balance over multiple blocks"""

    coroutines = []
    for block in blocks:
        coroutine = async_get_eth_balance(
            address=address, context=context, block=block, normalize=normalize
        )
        coroutines.append(coroutine)

    import asyncio

    return await asyncio.gather(*coroutines)


@typing.overload
async def async_get_eth_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    *,
    normalize: typing.Literal[False],
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> list[int]:
    ...


@typing.overload
async def async_get_eth_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: typing.Literal[True] = True,
    context: spec.Context = None,
) -> list[float]:
    ...


@typing.overload
async def async_get_eth_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    *,
    normalize: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Union[list[int], list[float]]:
    ...


async def async_get_eth_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    *,
    normalize: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Union[list[int], list[float]]:
    """get ETH balance of multiple addresses"""

    from ctc import rpc

    balances = await rpc.async_batch_eth_get_balance(
        addresses=addresses,
        context=context,
        block_number=block,
    )

    if normalize:
        balances = [balance / 1e18 for balance in balances]

    return balances



# from __future__ import annotations

# from ctc import evm
# from ctc import spec


# def get_sushiswap_factory(network: spec.NetworkReference) -> spec.Address:
#     network = evm.get_network_chain_id(network)
#     if network == 1:
#         return '0xc0aee478e3658e2610c5f7a4a2e1777ce9e4f2ac'
#     else:
#         raise Exception()



from __future__ import annotations

import typing
from typing_extensions import Literal, TypedDict
from . import abi_types


BinaryData = typing.Union[bytes, str]
GenericBinaryData = typing.Union[int, BinaryData]
# BlockSpec = typing.Union[
#     int, BinaryData, Literal['latest', 'earliest', 'pending']
# ]

BinaryFormat = Literal['binary', 'prefix_hex', 'raw_hex', 'integer']
ExtendedBinaryFormat = typing.Union[BinaryFormat, Literal['ascii']]

IntegerData = int

PrefixHexData = str
RawHexData = str
HexData = typing.Union[PrefixHexData, RawHexData]

Data = typing.Union[IntegerData, BinaryData, HexData]


# either a single datum or a triplet of (v, r, s)
Signature = typing.Union[Data, typing.Tuple[Data, Data, Data]]


class Eip712StructType(TypedDict):
    # implement nested structs once mypy#13297 is production ready
    # https://github.com/python/mypy/pull/13297
    name: str
    fields: typing.Mapping[str, abi_types.ABIDatatypeStr]



from __future__ import annotations

import os
import typing

from ctc import spec
from . import config_validate


def write_config_file(
    config_data: spec.Config,
    path: str,
    *,
    overwrite: typing.Literal[True, False, 'prompt'] = False,
    style: typing.Optional[str] = None,
    headless: bool = False,
) -> None:
    import json

    # validate config data
    config_validate.validate_config(config_data)

    # assert overwrite options
    if os.path.isfile(path):
        if overwrite is True:
            pass
        elif overwrite is False:
            raise Exception('use overwrite=True to overwrite an existing file')
        elif overwrite == 'prompt':
            import toolcli

            if not toolcli.input_yes_or_no(
                prompt='File already exists: '
                + str(path)
                + '\n\nOverwrite file?\n',
                style=style,
            ):
                raise Exception('must overwrite file to proceed')
        else:
            raise Exception('unknown value for overwrite: ' + str(overwrite))

    # write file
    directory = os.path.dirname(path)
    os.makedirs(directory, exist_ok=True)
    with open(path, 'w') as f:
        json.dump(config_data, f)




import pytest

from ctc import db


# test that every schema exists and is retrievable
@pytest.mark.parametrize('schema_name', db.get_schema_names())
def test_db_names_have_db_schemas(schema_name):
    db.get_raw_schema(schema_name)


@pytest.mark.parameterize('schema_name', db.get_schema_names())
def test_every_evm_schema_active_status_listed(schema_name):
    active_schemas = db.get_active_schemas()
    assert schema_name in active_schemas



from __future__ import annotations

import asyncio
import typing

import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc import evm
from ctc import spec

from ctc.protocols import ens_utils

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class ENSResult(TypedDict):
        address: spec.Address | None
        name: str | None
        owner: str | None
        expiration: int | None
        resolver: str | None


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_ens_command,
        'help': 'summarize ENS entry',
        'args': [
            {
                'name': 'name_or_address',
                'nargs': '+',
                'help': 'ENS name(s) or address(es)',
            },
            {'name': '--block', 'help': 'block number'},
            {
                'name': ['--verbose', '-v'],
                'help': 'display additional information',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045',
            'vitalik.eth',
            'vitalik.eth --verbose',
        ],
    }


async def async_ens_command(
    *,
    name_or_address: str,
    block: spec.BlockNumberReference,
    verbose: bool,
) -> None:
    if block is not None:
        block = evm.standardize_block_number(block)

    coroutines = [
        async_process_ens_arg(arg=arg, block=block) for arg in name_or_address
    ]
    results = await asyncio.gather(*coroutines)

    for r, result in enumerate(results):
        if r > 0:
            print()

        if result['name'] is None:
            toolstr.print_text_box(result['address'])
            print('[no ENS records]')
            continue
        elif result['address'] is None and result['owner'] is None:
            toolstr.print_text_box(result['name'])
            print('[no ENS records]')
            raise Exception()
            continue

        styles = cli.get_cli_styles()
        toolstr.print_text_box(result['name'], style=styles['title'])
        cli.print_bullet(key='address', value=result['address'])
        cli.print_bullet(key='owner', value=result['owner'])
        cli.print_bullet(key='resolver', value=result['resolver'])
        cli.print_bullet(
            key='namehash', value=ens_utils.hash_name(result['name'])
        )
        # print('- registered:', )
        cli.print_bullet(
            key='expiration',
            value=tooltime.timestamp_to_iso(result['expiration']).replace(
                'T', ' '
            ),
        )

        if verbose:
            text_records = await ens_utils.async_get_text_records(
                name=result['name']
            )
            if len(text_records) > 0:
                print()
                print()
                toolstr.print_header('Text Records', style=styles['title'])
                for key, value in sorted(text_records.items()):
                    cli.print_bullet(key=key, value=value)
            else:
                cli.print_bullet(value='no text records')


async def async_process_ens_arg(
    arg: str, block: spec.BlockNumberReference
) -> ENSResult:
    if '.' in arg:
        name = arg
        address = None
        address_coroutine = ens_utils.async_resolve_name(name, block=block)
    elif evm.is_address_str(arg):
        address = arg
        name = await ens_utils.async_reverse_lookup(address, block=block)
    else:
        raise Exception('could not parse inputs')

    if name == '':
        return {
            'address': address,
            'name': None,
            'owner': None,
            'expiration': None,
            'resolver': None,
        }

    owner_coroutine = ens_utils.async_get_owner(name=name)
    expiration_coroutine = ens_utils.async_get_expiration(name=name)
    resolver_coroutine = ens_utils.async_get_resolver(name=name)

    owner, expiration, resolver = await asyncio.gather(
        owner_coroutine,
        expiration_coroutine,
        resolver_coroutine,
    )

    if address is None:
        address = await address_coroutine

    return {
        'address': address,
        'name': name,
        'owner': owner,
        'expiration': expiration,
        'resolver': resolver,
    }




from __future__ import annotations

import typing
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict, Literal


if typing.TYPE_CHECKING:

    class GnosisSafeCreation(TypedDict):
        creation_block: int
        creation_transaction: spec.PrefixHexData
        factory: spec.Address
        implementation: spec.Address
        address: spec.Address
        version: Literal['1.1', '1.3']

    class SafeTransaction(TypedDict):
        to: spec.Address
        value: int
        data: bytes
        operation: int
        safeTxGas: int
        baseGas: int
        gasPrice: int
        gasToken: spec.Address
        refundReceiver: spec.Address
        nonce: int


safe_deployments_repository = 'https://github.com/safe-global/safe-deployments'

deployments = {
    1: {
        'factory__1.1.1': '0x76e2cfc1f5fa8f6a5b3fc4c8f4788f0116861f9b',
        'factory__1.3.0': '0xa6b71e26c5e0845f74c812102ca7114b6a896ab2',
        'safe__1.0.0': '0xb6029ea3b2c51d09a50b53ca8012feeb05bda35a',
        'safe__1.1.1': '0x34cfac646f301356faa8b21e94227e3583fe3f5f',
        'safe__1.2.0': '0x6851d6fdfafd08c0295c392436245e5bc78b0185',
        'safe__1.3.0': '0xd9db270c1b5e3bd161e8c8503c55ceabee709552',
        'safe_l2__1.3.0': '0x3e5c63644e683549055b9be8653de26e0b4cd36e',
    }
}

safe_transaction_keys = [
    'to',
    'value',
    'data',
    'operation',
    'safeTxGas',
    'baseGas',
    'gasPrice',
    'gasToken',
    'refundReceiver',
    'nonce',
]

safe_transaction_type: spec.Eip712StructType = {
    'name': 'SafeTx',
    'fields': {
        'to': 'address',
        'value': 'uint256',
        'data': 'bytes',
        'operation': 'uint8',
        'safeTxGas': 'uint256',
        'baseGas': 'uint256',
        'gasPrice': 'uint256',
        'gasToken': 'address',
        'refundReceiver': 'address',
        'nonce': 'uint256',
    },
}

safe_message_type = {
    'name': 'SafeMessage',
    'fields': {
        'message': 'bytes',
    },
}

function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'getOwners': {
        'inputs': [],
        'name': 'getOwners',
        'outputs': [
            {'internalType': 'address[]', 'name': '', 'type': 'address[]'}
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getThreshold': {
        'inputs': [],
        'name': 'getThreshold',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'nonce': {
        'inputs': [],
        'name': 'nonce',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'execTransaction': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'to',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'value',
                'type': 'uint256',
            },
            {
                'internalType': 'bytes',
                'name': 'data',
                'type': 'bytes',
            },
            {
                'internalType': 'enum Enum.Operation',
                'name': 'operation',
                'type': 'uint8',
            },
            {
                'internalType': 'uint256',
                'name': 'safeTxGas',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'baseGas',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'gasPrice',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'gasToken',
                'type': 'address',
            },
            {
                'internalType': 'address payable',
                'name': 'refundReceiver',
                'type': 'address',
            },
            {
                'internalType': 'bytes',
                'name': 'signatures',
                'type': 'bytes',
            },
        ],
        'name': 'execTransaction',
        'outputs': [
            {
                'internalType': 'bool',
                'name': 'success',
                'type': 'bool',
            },
        ],
        'stateMutability': 'payable',
        'type': 'function',
    },
}

event_abis: typing.Mapping[str, spec.EventABI] = {
    'AddedOwner': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'AddedOwner',
        'type': 'event',
    },
    'ChangedGuard': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'guard',
                'type': 'address',
            },
        ],
        'name': 'ChangedGuard',
        'type': 'event',
    },
    'ChangedThreshold': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'threshold',
                'type': 'uint256',
            },
        ],
        'name': 'ChangedThreshold',
        'type': 'event',
    },
    'ExecutionSuccess': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'bytes32',
                'name': 'txHash',
                'type': 'bytes32',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'payment',
                'type': 'uint256',
            },
        ],
        'name': 'ExecutionSuccess',
        'type': 'event',
    },
    'RemovedOwner': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'RemovedOwner',
        'type': 'event',
    },
    'SafeSetup': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'initiator',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address[]',
                'name': 'owners',
                'type': 'address[]',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'threshold',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'initializer',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'fallbackHandler',
                'type': 'address',
            },
        ],
        'name': 'SafeSetup',
        'type': 'event',
    },
}



from __future__ import annotations

import pytest

import ctc


example_deposits = [
    {
        'kwargs': {
            'token': '0xf9a98a9452485ed55cd3ce5260c2b71c9807b11a',
            'start_block': 15060729,
            'end_block': 15755513,
        },
        'n_results': 49,
    },
]

example_withdraws = [
    {
        'kwargs': {
            'token': '0xf9a98a9452485ed55cd3ce5260c2b71c9807b11a',
            'start_block': 15060729,
            'end_block': 15755513,
        },
        'n_results': 19,
    },
]


@pytest.mark.parametrize('example', example_deposits)
async def test_get_erc4626_deposits(example):
    deposits = await ctc.async_get_erc4626_deposits(**example['kwargs'])
    assert len(deposits) == example['n_results']


@pytest.mark.parametrize('example', example_withdraws)
async def test_get_erc4626_withdraws(example):
    withdraws = await ctc.async_get_erc4626_withdraws(**example['kwargs'])
    assert len(withdraws) == example['n_results']



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from .. import chainlink_spec
from .. import chainlink_feed_metadata
from . import feed_data

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_composite_feed_data(
    composite_feed: typing.Sequence[chainlink_spec._FeedReference],
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    invert: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:
    # TODO: other ways of specifying composites
    import asyncio
    import polars as pl

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    # queue requests
    coroutines = []
    for feed in composite_feed:
        coroutine = feed_data.async_get_feed_data(
            feed,
            start_block=start_block,
            end_block=end_block,
            normalize=False,
            invert=False,
            interpolate=True,
            fields='answer',
            context=context,
        )
        coroutines.append(coroutine)

    datas: typing.Sequence[spec.DataFrame] = await asyncio.gather(*coroutines)

    # compute product
    product = datas[0] * datas[1]
    for data in datas[2:]:
        product = product * data

    # normalize
    feeds_decimals_coroutines = [
        chainlink_feed_metadata.async_get_feed_decimals(
            feed=feed,
            context=context,
        )
        for feed in composite_feed
    ]
    feeds_decimals = await asyncio.gather(*feeds_decimals_coroutines)
    composite_decimals = sum(feeds_decimals)
    product = product / (10 ** composite_decimals)

    # invert
    if invert:
        product = product.with_columns(1 / pl.col('answer'))

    return product



from __future__ import annotations

from ctc import rpc
from ctc import spec
from ctc.protocols import chainlink_utils

from ... import rari_abis
from . import token_price


async def async_get_total_borrowed(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> int:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['totalBorrows'],
        empty_token=0,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_total_liquidity(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> int:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['getCash'],
        empty_token=0,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_reserves(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> int:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['totalReserves'],
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_ctoken_utilization(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> float:
    borrowed_coroutine = async_get_total_borrowed(ctoken, block)
    liquidity_coroutine = async_get_total_liquidity(ctoken, block)

    borrowed = await borrowed_coroutine
    liquidity = await liquidity_coroutine

    total = borrowed + liquidity

    if total == 0:
        return 0
    else:
        return borrowed / total


async def async_get_ctoken_tvl_and_tvb(
    ctoken: spec.Address,
    *,
    oracle: spec.Address | None = None,
    eth_price: spec.Number | None = None,
    block: spec.BlockNumberReference = 'latest',
    in_usd: bool = True,
) -> dict[str, spec.Number]:
    """combined into one function because tvl requires both to compute anyway"""
    import asyncio

    if not in_usd:
        borrowed_coroutine = asyncio.create_task(
            async_get_total_borrowed(ctoken=ctoken, block=block)
        )
        liquidity_coroutine = asyncio.create_task(
            async_get_total_liquidity(ctoken=ctoken, block=block)
        )

        borrowed = (await borrowed_coroutine) / 1e18
        liquidity = (await liquidity_coroutine) / 1e18

        return {'tvb': borrowed, 'tvl': borrowed + liquidity}

    else:

        if oracle is None:
            oracle = await token_price._async_get_ctoken_oracle(ctoken)

        if eth_price is None:
            eth_price = await chainlink_utils.async_get_eth_price(block=block)

        # send queries
        borrowed_coroutine = asyncio.create_task(
            async_get_total_borrowed(ctoken=ctoken, block=block)
        )
        liquidity_coroutine = asyncio.create_task(
            async_get_total_liquidity(ctoken=ctoken, block=block)
        )
        price_coroutine = asyncio.create_task(
            token_price.async_get_ctoken_price(
                ctoken=ctoken, oracle=oracle, block=block
            )
        )

        # receive results
        borrowed = (await borrowed_coroutine) / 1e18
        liquidity = (await liquidity_coroutine) / 1e18
        price = await price_coroutine

        # compute output
        return {
            'tvb': borrowed * price * eth_price,
            'tvl': (borrowed + liquidity) * price * eth_price,
        }



from __future__ import annotations

import asyncio
import typing

from ctc import rpc
from ctc import spec


def get_contract_type_guards() -> dict[
    str,
    typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, bool]],
]:
    return {
        'chainlink_feed': async_is_chainlink_feed,
        'uniswap_v2_pool': async_is_uniswap_v2_pool,
        'uniswap_v3_pool': async_is_uniswap_v3_pool,
        'curve_pool': async_is_curve_pool,
        'sushi_pool': async_is_sushi_pool,
        'balancer_v2_pool': async_is_balancer_v2_pool,
    }


async def async_get_contract_type(
    address: spec.Address,
    context: spec.Context = None,
) -> str:

    contract_guards = get_contract_type_guards()
    coroutines = [
        contract_guard(address=address, context=context)
        for contract_guard in contract_guards.values()
    ]
    tasks = [asyncio.create_task(coroutine) for coroutine in coroutines]
    contract_guard_names = list(contract_guards.keys())
    while True:
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in done:
            result = task.result()
            if result:
                return contract_guard_names[tasks.index(task)]
        if len(pending) == 0:
            raise Exception('could not determine contract type')


#
# # specific guard
#


async def async_is_chainlink_feed(
    address: spec.Address, context: spec.Context = None
) -> bool:
    try:
        await rpc.async_eth_call(
            to_address=address,
            function_abi={'name': 'aggregator', 'inputs': [], 'outputs': []},
            context=context,
        )
        return True
    except spec.RpcException:
        return False


async def async_is_uniswap_v2_pool(
    address: spec.Address, context: spec.Context = None
) -> bool:
    try:
        factory = await rpc.async_eth_call(
            to_address=address,
            function_abi={
                'name': 'factory',
                'inputs': [],
                'outputs': [{'type': 'address'}],
            },
            context=context,
        )
        return bool(factory == '0x5c69bee701ef814a2b6a3edd4b1652cb9cc5aa6f')
    except spec.RpcException:
        return False


async def async_is_uniswap_v3_pool(
    address: spec.Address, context: spec.Context = None
) -> bool:
    try:
        factory = await rpc.async_eth_call(
            to_address=address,
            function_abi={
                'name': 'factory',
                'inputs': [],
                'outputs': [{'type': 'address'}],
            },
            context=context,
        )
        return bool(factory == '0x1f98431c8ad98523631ae4a59f267346ea31f984')
    except spec.RpcException:
        return False


async def async_is_curve_pool(
    address: spec.Address, context: spec.Context = None
) -> bool:
    # doesn't work for early pools
    try:
        await rpc.async_eth_call(
            to_address=address,
            function_abi={'name': 'A', 'inputs': [], 'outputs': []},
            context=context,
        )
        return True
    except spec.RpcException:
        return False


async def async_is_sushi_pool(
    address: spec.Address, context: spec.Context = None
) -> bool:
    try:
        factory = await rpc.async_eth_call(
            to_address=address,
            function_abi={
                'name': 'factory',
                'inputs': [],
                'outputs': [{'type': 'address'}],
            },
            context=context,
        )
        return bool(factory == '0xc0aee478e3658e2610c5f7a4a2e1777ce9e4f2ac')
    except spec.RpcException:
        return False


async def async_is_balancer_v2_pool(
    address: spec.Address, context: spec.Context = None
) -> bool:
    try:
        await rpc.async_eth_call(
            to_address=address,
            function_abi={'name': 'getPoolId', 'inputs': [], 'outputs': []},
            context=context,
        )
        return True
    except spec.RpcException:
        return False




from __future__ import annotations

import toolcli
import toolstr

from ctc.protocols import coingecko_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_cg_command,
        'help': 'output coingecko market data',
        'args': [
            {
                'name': 'tokens',
                'help': 'token to display information of',
                'nargs': '*',
            },
            {
                'name': ['-t', '--time'],
                'help': 'time of chart (e.g. 7d, 1y, max)',
                'dest': 'timelength',
            },
            {'name': '-n', 'help': 'number of entries to include in output'},
            {
                'name': ['--verbose', '-v'],
                'action': 'store_const',
                'const': True,
                'default': None,
                'help': 'include extra data',
            },
            {'name': '--height', 'help': 'height, number of rows per asset'},
            {'name': '--width', 'help': 'width of sparklines'},
            {
                'name': '--update',
                'help': 'update stored coingecko token db',
                'action': 'store_true',
            },
        ],
        'examples': {
            '': {'runnable': False},
            'CRV -t 100d': {'runnable': False},
            'ETH / BTC': {'runnable': False},
            'LINK / ETH -t 24h': {'runnable': False},
            '--update': {'runnable': False},
        },
    }


async def async_cg_command(
    *,
    tokens: str,
    n: int,
    verbose: bool,
    height: int | None,
    width: str | int | None,
    update: bool,
    timelength: str | None,
) -> None:

    if update:
        await coingecko_utils.async_get_token_list(
            context={'cache': False}, update=True
        )
        return

    if height is None:
        height = 1
    height = int(height)

    if isinstance(width, str):
        width = int(width)

    if timelength is not None:
        if timelength == 'max':
            days = None
        else:
            import tooltime

            days = round(tooltime.timelength_to_seconds(timelength) / 86400)
    else:
        days = 30

    if len(tokens) == 0:

        terminal_rows = toolcli.get_n_terminal_rows()
        terminal_columns = toolcli.get_n_terminal_cols()
        if n is None:
            if terminal_columns > 172:
                n_columns = 2
                gap = 4
                n = (terminal_rows - 4) * 2

            else:
                n_columns = None
                gap = None

                n = terminal_rows - 4
                n = int(n / height)

        else:
            n = int(n)

            if n > terminal_rows - 4 and terminal_columns > 172:
                n_columns = 2
                gap = 4

        data = await coingecko_utils.async_get_market_data(n)

        if verbose is None:
            verbose = toolcli.get_n_terminal_cols() >= 96

        coingecko_utils.print_market_data(
            data=data,
            verbose=verbose,
            height=height,
            width=width,
            n_columns=n_columns,
            gap=gap,
        )

        note = 'coingecko data @ ' + str(data[0]['last_updated'])
        toolstr.print(note, style='#555555')

    elif len(tokens) == 3 and tokens[1] == '/':
        await coingecko_utils.async_print_coin_quotient_summary(
            tokens[0],
            tokens[2],
            days=days,
        )

    elif len(tokens) == 1:

        await coingecko_utils.async_print_token_data_summary(
            query=tokens[0],
            verbose=verbose,
            update=update,
            days=days,
        )

    else:
        raise Exception('could not parse inputs, use --help for details')




from .transaction_convert import *
from .transaction_crud import *
from .transaction_hashes import *
from .transaction_serialize import *
from .transaction_signatures import *
from .transaction_summary import *
from .transaction_types import *



from __future__ import annotations

import typing

from ctc import spec
from .. import abi_coding_utils
from . import event_abi_parsing
from . import event_abi_queries

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

    EventABIList = typing.Sequence[spec.EventABI]
    EventABIMap = typing.Mapping[str, spec.EventABI]
    ColumnPrefixType = Literal['arg', 'event_name', 'event_hash']


async def async_decode_events_dataframe(
    events: spec.DataFrame,
    event_abis: EventABIList | EventABIMap | None = None,
    *,
    context: spec.Context,
    column_prefix_type: ColumnPrefixType | None = None,
    column_prefix: str | None = None,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    integer_output_format: spec.IntegerOutputFormat | None = None,
    convert_invalid_str_to_none: bool = False,
    convert_invalid_str_to: None | str = None,
) -> spec.DataFrame:
    """decode a dataframe that contains raw logs

    TODO: specifically handle logs that do not provide event topics

    columns are prefixed according to input settings
    - column_prefix_type == 'arg' --> use 'arg__' as prefix
    - column_prefix_type == 'event_name' --> use '{event_name}__' as prefix
    - column_prefix_type == 'event_hash' --> use '{event_hash}__' as prefix
    - column_prefix_type == None --> use column_prefix if not None, otherwise ''
    """

    import polars as pl
    from ctc.toolbox import pl_utils

    # handle case of no events
    if len(events) == 0:
        _, df_schema, _ = _create_event_arg_schema(
            event_abis=event_abis,
            column_prefix_type=column_prefix_type,
            column_prefix=column_prefix,
            integer_output_format=integer_output_format,
        )
        arg_df = pl.DataFrame(schema=df_schema)  # type: ignore
        return pl.concat([events, arg_df], how='horizontal')

    # get event abis
    event_abis = await _async_get_events_abis(
        events=events, event_abis=event_abis, context=context
    )

    # create dataframe schema
    event_schemas, df_schema, event_hash_offsets = _create_event_arg_schema(
        event_abis=event_abis,
        column_prefix_type=column_prefix_type,
        column_prefix=column_prefix,
        integer_output_format=integer_output_format,
    )

    # create iterators for relevant columns
    n_decode_topics = max(
        len(event_schema['indexed_types'])
        for event_schema in event_schemas.values()
    )
    if n_decode_topics >= 1:
        topic1 = events['topic1'].to_list()
    else:
        topic1 = None
    if n_decode_topics >= 2:
        topic2 = events['topic2'].to_list()
    else:
        topic2 = None
    if n_decode_topics >= 3:
        topic3 = events['topic3'].to_list()
    else:
        topic3 = None
    decode_unindexed = any(
        len(event_schema['unindexed_types']) > 0
        for event_schema in event_schemas.values()
    )
    if decode_unindexed:
        unindexed = events['unindexed'].to_list()
    topic_iterators = [topic1, topic2, topic3]

    # decode
    decoded_events: typing.Sequence[typing.MutableSequence[typing.Any]]
    decoded_events = [[] for i in range(len(df_schema))]
    for e, event_hash in enumerate(events['event_hash'].to_list()):
        i = event_hash_offsets[event_hash]
        event_schema = event_schemas[event_hash]

        for c in range(i):
            decoded_events[c].append(None)

        # decode indexed data
        for indexed_type, indexed_name, topic_iterator in zip(
            event_schema['indexed_types'],
            event_schema['indexed_names'],
            topic_iterators,
        ):
            if topic_iterator is None:
                raise Exception('topic_iterator not set')
            if (
                indexed_type in ['bytes', 'string']
                or indexed_type.endswith(']')
                or indexed_type.endswith(')')
            ):
                value = topic_iterator[e]
            else:
                raw_value = topic_iterator[e]
                if raw_value is not None:
                    value = abi_coding_utils.abi_decode(
                        topic_iterator[e],
                        indexed_type,
                        convert_invalid_str_to=convert_invalid_str_to,
                        convert_invalid_str_to_none=convert_invalid_str_to_none,
                    )
                else:
                    value = raw_value
            decoded_events[i].append(value)
            i = i + 1

        # decode unindexed data
        if len(event_schema['unindexed_types']) > 0:
            # unindexed_decoded = abi_coding_utils.abi_decode(
            #     unindexed[e],
            #     event_schema['unindexed_types'],
            #     convert_invalid_str_to=convert_invalid_str_to,
            #     convert_invalid_str_to_none=convert_invalid_str_to_none,
            # )
            if len(event_schema['unindexed_types']) == 1:
                result = abi_coding_utils.abi_decode(
                    unindexed[e],
                    event_schema['unindexed_types'][0],
                    convert_invalid_str_to=convert_invalid_str_to,
                    convert_invalid_str_to_none=convert_invalid_str_to_none,
                )
                unindexed_decoded = [result]
            else:
                unindexed_decoded = abi_coding_utils.abi_decode(
                    unindexed[e],
                    event_schema['unindexed_types'],
                    convert_invalid_str_to=convert_invalid_str_to,
                    convert_invalid_str_to_none=convert_invalid_str_to_none,
                )
            for value in unindexed_decoded:
                decoded_events[i].append(value)
                i = i + 1

        for c in range(i, len(df_schema)):
            decoded_events[c].append(None)

    # convert decimals (can remove if polars allows direct creation in future)
    new_columns = []
    for column, (column_name, dtype) in zip(decoded_events, df_schema):
        if dtype == pl.Decimal:
            import decimal

            new_columns.append([decimal.Decimal(value) for value in column])
        else:
            new_columns.append(column)
    decoded_events = new_columns

    # convert to dataframe
    decoded = pl.DataFrame(decoded_events, schema=df_schema, orient='col')  # type: ignore

    # convert binary columns to hex
    if binary_output_format == 'prefix_hex':
        decoded = pl_utils.binary_columns_to_prefix_hex(decoded)

    return decoded


def _create_event_arg_schema(
    *,
    event_abis: EventABIList | EventABIMap | None = None,
    column_prefix_type: ColumnPrefixType | None = None,
    column_prefix: str | None = None,
    integer_output_format: spec.IntegerOutputFormat | None = None,
) -> tuple[
    typing.Mapping[str, spec.EventSchema],
    typing.Sequence[tuple[str, spec.IntegerOutputFormat]],
    typing.Mapping[str, int],
]:
    """does not take binary_output_format into account, that happens later"""

    # format event_abi's as dict
    if not isinstance(event_abis, dict):
        if isinstance(event_abis, (list, tuple)):
            event_abis = {
                event_abi_parsing.get_event_hash(event_abi): event_abi
                for event_abi in event_abis
            }
        elif event_abis is None:
            event_abis = {}
        else:
            raise Exception('unknown events format')

    # get event schemas
    event_schemas: typing.Mapping[str, spec.EventSchema] = {
        event_hash: event_abi_parsing.get_event_schema(event_abi)
        for event_hash, event_abi in event_abis.items()
    }

    # get column prefix
    if column_prefix_type is None and column_prefix is None:
        if len(event_abis) == 1:
            column_prefix_type = 'arg'
        else:
            event_names = {
                abi.get('name')
                for abi in event_abis.values()
                if abi.get('name') is not None
            }
            if len(event_names) == len(event_abis):
                column_prefix_type = 'event_name'
            else:
                column_prefix_type = 'event_hash'
    if column_prefix_type == 'arg':
        column_prefix = 'arg__'
    elif column_prefix_type == 'event_name':
        column_prefix = '{event_name}__'
    elif column_prefix_type == 'event_hash':
        column_prefix = '{event_hash}__'
    else:
        if column_prefix is None:
            column_prefix = ''

    # create dataframe schema
    df_schema = []
    used_names = []
    offsets = [0]
    for event_hash in event_schemas.keys():
        event_schema = event_schemas[event_hash]
        if '{event_name}' in column_prefix:
            if 'name' in event_abis[event_hash]:
                event_name = event_abis[event_hash]['name']
            else:
                raise Exception('event name not specified in event abi')
            used_column_prefix = column_prefix.format(
                event_hash=event_hash,
                event_name=event_name,
            )
        else:
            used_column_prefix = column_prefix.format(event_hash=event_hash)
        for abi_type, name in zip(event_schema['types'], event_schema['names']):
            pl_type = _abi_type_to_polars_dtype(
                abi_type=abi_type,
                name=name,
                integer_output_format=integer_output_format,
            )
            name = used_column_prefix + name
            if name in used_names:
                raise Exception('naming conflict ' + str(name))

            df_schema.append((name, pl_type))
            used_names.append(name)
        offsets.append(len(df_schema))
    event_hash_offsets = dict(zip(event_schemas.keys(), offsets))

    return event_schemas, df_schema, event_hash_offsets


async def _async_get_events_abis(
    events: spec.DataFrame,
    event_abis: EventABIList | EventABIMap | None = None,
    *,
    context: spec.Context,
) -> typing.Mapping[str, spec.EventABI]:
    import polars as pl

    # package input abi's
    if isinstance(event_abis, dict):
        abis: typing.MutableMapping[str, spec.EventABI] = event_abis
    elif event_abis is None:
        abis = {}
    elif isinstance(event_abis, list):
        abis = {}
        for event_abi in event_abis:
            event_hash = event_abi_parsing.get_event_hash(event_abi)
            abis[event_hash] = event_abi
    else:
        raise Exception()

    # acquire missing event abi's
    event_hashes = list(events['event_hash'].unique())
    missing_event_types = set(event_hashes) - set(abis.keys())
    if len(missing_event_types) > 0:
        import asyncio

        event_types = events.groupby('event_hash').agg(
            pl.col('contract_address').unique()
        )
        coroutines = []
        for event_hash, contract_addresses in event_types.rows():
            coroutine = event_abi_queries.async_get_event_abi(
                event_hash=event_hash,
                contract_addresses=contract_addresses,
                context=context,
            )
            coroutines.append(coroutine)
        new_event_abis = await asyncio.gather(*coroutines)
        abis.update(
            dict(zip(event_types['event_hash'].to_list(), new_event_abis))
        )

    # remove unused abi's
    abis = {k: v for k, v in abis.items() if k in event_hashes}

    return abis


def _abi_type_to_polars_dtype(
    *,
    abi_type: str,
    name: str,
    integer_output_format: spec.IntegerOutputFormat | None = None,
) -> type[object]:
    import polars as pl

    if (
        abi_type.startswith('int') or abi_type.startswith('uint')
    ) and integer_output_format is not None:
        # if a dict, using integer_output_format is optional. otherwise must use
        if isinstance(integer_output_format, dict):
            if name in integer_output_format:
                return integer_output_format[name]  # type: ignore
        else:
            if integer_output_format in [int, 'integer']:
                return int
            elif integer_output_format in [object, 'object']:
                return pl.Object
            elif integer_output_format in [float, 'float']:
                return pl.Float64
            elif isinstance(integer_output_format, pl.datatypes.DataTypeClass):
                return integer_output_format
            elif (
                integer_output_format == 'decimal'
                or str(type(integer_output_format)) == "<class 'decimal.Decimal'>"
            ):
                return pl.Decimal
            else:
                raise Exception(
                    'invalid integer_output_format: ' + str(integer_output_format)
                )

    if abi_type == 'bool':
        return pl.datatypes.Boolean
    elif abi_type == 'address':
        return pl.datatypes.Utf8
    elif abi_type == 'string':
        return pl.datatypes.Utf8
    elif abi_type in ['bytes', 'bytes8', 'bytes16', 'bytes24', 'bytes32']:
        return pl.datatypes.Binary
    elif abi_type == 'function':
        return pl.datatypes.Object
    elif abi_type in [
        'int8',
        'int16',
        'int24',
        'int32',
        'int40',
        'int48',
        'int56',
        'int64',
        'uint8',
        'uint16',
        'uint24',
        'uint32',
    ]:
        return pl.datatypes.Int64
    elif abi_type in [
        'int72',
        'int80',
        'int88',
        'int96',
        'int104',
        'int112',
        'int120',
        'int128',
        'int136',
        'int144',
        'int152',
        'int160',
        'int168',
        'int176',
        'int184',
        'int192',
        'int200',
        'int208',
        'int216',
        'int224',
        'int232',
        'int240',
        'int248',
        'int256',
        'uint72',
        'uint80',
        'uint88',
        'uint96',
        'uint104',
        'uint112',
        'uint120',
        'uint128',
        'uint136',
        'uint144',
        'uint152',
        'uint160',
        'uint168',
        'uint176',
        'uint184',
        'uint192',
        'uint200',
        'uint208',
        'uint216',
        'uint224',
        'uint232',
        'uint240',
        'uint248',
        'uint256',
    ]:
        return pl.datatypes.Object
    elif (
        abi_type.startswith('fixed') or abi_type.startswith('ufixed')
    ) and not abi_type.endswith(']'):
        return pl.datatypes.Decimal
    elif abi_type.endswith(']') or abi_type.endswith(')'):
        return pl.datatypes.Object
    else:
        raise Exception('unknown abi type')




from __future__ import annotations

import asyncio
import typing

import toolcli
import tooltime

from ctc import evm
from ctc import spec
from ctc.protocols import aave_v2_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_aave_command,
        'help': 'output aave usage statistics',
        'args': [
            {
                'name': 'token',
                'help': 'show information related to this token',
                'nargs': '?',
            },
            {
                'name': ['--verbose', '-v'],
                'help': 'show additional addresses',
                'action': 'store_true',
            },
            {
                'name': '--block',
                'help': 'block to get addresses of',
            },
            {
                'name': '--time',
                'help': 'timescale to use for token-specific charts',
            },
            {
                'name': '-n',
                'help': 'number of blocks to sample',
                'default': 40,
                'type': int,
            },
        ],
        'examples': [
            '',
            '--verbose',
            '--block 14000000',
            'DAI',
            'DAI --time 1y -n 10',
            'DAI --time 1y --verbose -n 10',
        ],
    }


async def async_aave_command(
    *,
    token: str | None,
    verbose: bool,
    block: str | None,
    time: str | None,
    n: int | None,
) -> None:

    if block is None:
        block = 'latest'
    block_reference = evm.standardize_block_number(block)

    if token is not None:

        blocks = await _async_sample_blocks(
            timelength=time,
            n_samples=n,
            end_block=block_reference,
        )

        await aave_v2_utils.async_print_token_market_summary(
            token=token,
            verbose=verbose,
            blocks=blocks,
        )

    else:
        await aave_v2_utils.async_print_token_markets_summary(
            verbose=verbose,
            block=block,
        )


async def _async_sample_blocks(
    *,
    n_samples: int | None,
    timelength: tooltime.Timelength | None = None,
    end_block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:

    import time
    import numpy as np

    if n_samples is None:
        n_samples = 100

    if timelength is None:
        timelength = '30d'

    if end_block is None or end_block == 'latest':
        end_block_task = asyncio.create_task(
            evm.async_get_latest_block_number()
        )

    # compute timelength
    end_time = int(time.time())
    start_time = end_time - tooltime.timelength_to_seconds(timelength)
    start_block = await evm.async_get_block_of_timestamp(
        start_time,
        mode='<=',
    )

    if end_block is None or end_block == 'latest':
        end_block = await end_block_task
    if not isinstance(end_block, int):
        raise Exception('invalid type for end_block')

    samples = np.linspace(start_block, end_block, n_samples).astype(int)

    return samples  # type: ignore



from __future__ import annotations

import typing

import toolsql

from ctc import config
from .. import market_utils
from . import coingecko_schema_defs
from . import coingecko_statements


async def async_intake_tokens(
    tokens: typing.Sequence[coingecko_schema_defs.CoingeckoToken],
    n_ranked: int = 500,
) -> None:

    # add rank data
    market_data = await market_utils.async_get_market_data(n_ranked)
    ranks_of_symbols = {
        item['id']: item['market_cap_rank'] for item in market_data
    }
    for token in tokens:
        rank = ranks_of_symbols.get(token['id'])
        if rank is not None:
            token['market_cap_rank'] = rank

    # add to db
    db_config = config.get_context_db_config(
        schema_name='coingecko',
        context={},
    )
    async with toolsql.async_connect(db_config) as conn:
        await coingecko_statements.async_upsert_tokens(
            tokens=tokens,
            conn=conn,
        )




from __future__ import annotations

import toolsql


_event_binary_fields = [
    'contract_address',
    'transaction_hash',
    'event_hash',
    'topic1',
    'topic2',
    'topic3',
    'unindexed',
]

_event_query_binary_fields = [
    'contract_address',
    'event_hash',
    'topic1',
    'topic2',
    'topic3',
]

events_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'events': {
            'columns': [
                {
                    'name': 'block_number',
                    'type': 'Integer',
                    'primary': True,
                    'nullable': False,
                },
                {
                    'name': 'transaction_index',
                    'type': 'Integer',
                    'primary': True,
                    'nullable': False,
                },
                {
                    'name': 'log_index',
                    'type': 'Integer',
                    'primary': True,
                    'nullable': False,
                },
                {
                    'name': 'transaction_hash',
                    'type': 'Binary',
                    'nullable': False,
                },
                {'name': 'contract_address', 'type': 'Binary', 'index': True},
                {'name': 'event_hash', 'type': 'Binary', 'index': True},
                {'name': 'topic1', 'type': 'Binary', 'index': True},
                {'name': 'topic2', 'type': 'Binary', 'index': True},
                {'name': 'topic3', 'type': 'Binary', 'index': True},
                {'name': 'unindexed', 'type': 'Binary'},
                {'name': 'decoded', 'type': 'JSON'},
            ],
        },
        'event_queries': {
            'columns': [
                {
                    'name': 'query_id',
                    'type': 'Integer',
                    'primary': True,
                    'autoincrement': True,
                },
                {'name': 'contract_address', 'type': 'Binary', 'index': True},
                {
                    'name': 'start_block',
                    'type': 'Integer',
                    'index': True,
                    'nullable': False,
                },
                {
                    'name': 'end_block',
                    'type': 'Integer',
                    'index': True,
                    'nullable': False,
                },
                {'name': 'event_hash', 'type': 'Binary', 'index': True},
                {'name': 'topic1', 'type': 'Binary', 'index': True},
                {'name': 'topic2', 'type': 'Binary', 'index': True},
                {'name': 'topic3', 'type': 'Binary', 'index': True},
                {
                    'name': 'query_type',
                    'type': 'Integer',
                    'index': True,
                    'nullable': False,
                },
            ],
        },
    },
}




from __future__ import annotations

import typing

from ctc import spec
from .. import abi_utils
from .. import address_utils
from .. import binary_utils
from . import erc20_generic
from . import erc20_spec


async def async_get_erc20_address(
    token: spec.ERC20Reference,
    *,
    context: spec.Context = None,
    case_insensitive_query: bool = True,
) -> spec.ERC20Address:
    """return address of input token, input as either symbol or address"""

    if isinstance(token, bytes):
        token = binary_utils.to_hex(token)

    if address_utils.is_address_str(token):
        return token
    elif isinstance(token, str):
        from ctc import db

        metadata = await db.async_query_erc20_metadata(
            symbol=token,
            context=context,
        )

        # if no result, try again case insensitive
        if metadata is None:
            metadata = await db.async_query_erc20_metadata(
                symbol=token,
                context=context,
                case_insensitive_symbol=case_insensitive_query,
            )

        if metadata is not None:
            address = metadata['address']
            if isinstance(address, str):
                return address

    raise Exception('could not get token address')


async def async_is_erc20(
    address_or_abi: spec.Address | spec.ContractABI,
    *,
    context: spec.Context = None,
) -> bool:
    """return whether an address implements ERC20 spec in its abi"""

    # get contract abi
    if address_utils.is_address_str(address_or_abi):
        contract_abi = await abi_utils.async_get_contract_abi(
            address_or_abi, context=context
        )
    elif isinstance(address_or_abi, list):
        contract_abi = address_or_abi
    contract_abi_by_selectors = abi_utils.get_contract_abi_by_selectors(
        contract_abi
    )

    # get erc20 abi
    erc20_abi: spec.ContractABI = list(erc20_spec.erc20_function_abis.values())
    erc20_abi += list(erc20_spec.erc20_event_abis.values())
    erc20_abi_by_selectors = abi_utils.get_contract_abi_by_selectors(erc20_abi)

    # compare abis
    for selector in erc20_abi_by_selectors.keys():
        if selector not in contract_abi_by_selectors:
            return False
    return True


async def async_get_default_erc20_tokens(
    context: spec.Context = None,
) -> typing.Sequence[spec.ERC20Metadata]:
    """get list of default ERC20s

    TODO: add db table for tracking erc20 token lists
    - use a token list as tokens to check during block exploration
    - "default" can be one of the token lists
    """
    from ctc.config.setup_utils.default_data import default_erc20s

    return default_erc20s.load_default_erc20s(context=context)


async def async_get_erc20_metadata(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> spec.ERC20Metadata:
    """get metadata for ERC20 token"""

    address = await async_get_erc20_address(token, context=context)

    symbol_coroutine = async_get_erc20_symbol(
        token=token, block=block, context=context, **rpc_kwargs
    )
    decimals_coroutine = async_get_erc20_decimals(
        token=token, block=block, context=context, **rpc_kwargs
    )
    name_coroutine = async_get_erc20_name(
        token=token, block=block, context=context, **rpc_kwargs
    )

    import asyncio

    symbol, decimals, name = await asyncio.gather(
        symbol_coroutine,
        decimals_coroutine,
        name_coroutine,
    )

    return {
        'address': address,
        'symbol': symbol,
        'decimals': decimals,
        'name': name,
    }


#
# # decimals
#


async def async_get_erc20_decimals(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> int:
    """get decimals of an erc20"""

    from ctc import config

    result: spec.ERC20Metadata | None = None

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='erc20_metadata', context=context
    )

    if read_cache:
        from ctc import db

        token = await async_get_erc20_address(token, context=context)
        result = await db.async_query_erc20_metadata(
            address=token, context=context
        )
        if result is not None and result['decimals'] is not None:
            return result['decimals']

    decimals_result = await erc20_generic.async_erc20_eth_call(
        function_name='decimals',
        token=token,
        block=block,
        context=context,
        **rpc_kwargs,
    )
    if not isinstance(decimals_result, int) and not rpc_kwargs.get(
        'convert_reverts_to_none'
    ):
        raise Exception('invalid rpc result')
    decimals = decimals_result

    if write_cache:
        if result is not None:
            result['decimals'] = decimals
        else:
            data = {'address': token, 'decimals': decimals}
            if typing.TYPE_CHECKING:
                result = typing.cast(spec.ERC20Metadata, data)
            else:
                result = data
        await db.async_intake_erc20_metadata(context=context, **result)

    return decimals


async def async_get_erc20s_decimals(
    tokens: typing.Iterable[spec.ERC20Reference],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[int]:
    """get decimals of multiple erc20s"""
    return await erc20_generic.async_erc20s_eth_calls(
        function_name='decimals',
        tokens=tokens,
        block=block,
        context=context,
        **rpc_kwargs,
    )


async def async_get_erc20_decimals_by_block(
    token: spec.ERC20Reference,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[int]:
    """get decimals of an erc20 across multiple blocks"""
    return await erc20_generic.async_erc20_eth_call_by_block(
        function_name='decimals',
        token=token,
        blocks=blocks,
        context=context,
        **rpc_kwargs,
    )


#
# # name
#


@typing.overload
async def async_get_erc20_name(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to_none: typing.Literal[True],
    **rpc_kwargs: typing.Any,
) -> str | None:
    ...


@typing.overload
async def async_get_erc20_name(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to_none: typing.Literal[False] = False,
    **rpc_kwargs: typing.Any,
) -> str:
    ...


@typing.overload
async def async_get_erc20_name(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to_none: bool,
    **rpc_kwargs: typing.Any,
) -> str | None:
    ...


async def async_get_erc20_name(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to_none: bool = False,
    **rpc_kwargs: typing.Any,
) -> str | None:
    """get name of an erc20"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='erc20_metadata', context=context
    )

    result: spec.ERC20Metadata | None = None
    if read_cache:
        from ctc import db

        token = await async_get_erc20_address(token, context=context)
        result = await db.async_query_erc20_metadata(
            address=token, context=context
        )
        if result is not None and result['name'] is not None:
            return result['name']

    rpc_result: str | None = await erc20_generic.async_erc20_eth_call(
        function_name='name',
        token=token,
        block=block,
        context=context,
        **rpc_kwargs,
    )
    if not isinstance(rpc_result, str) and not convert_reverts_to_none:
        raise Exception('invalid rpc result')
    name = rpc_result

    if write_cache:
        if result is not None:
            result['name'] = name
        else:
            data = {'address': token, 'name': name}
            if typing.TYPE_CHECKING:
                result = typing.cast(spec.ERC20Metadata, data)
            else:
                result = data

        await db.async_intake_erc20_metadata(context=context, **result)

    return name


async def async_get_erc20s_names(
    tokens: typing.Iterable[spec.ERC20Reference],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[str]:
    """get name of multiple erc20s"""
    return await erc20_generic.async_erc20s_eth_calls(
        function_name='name',
        tokens=tokens,
        block=block,
        context=context,
        **rpc_kwargs,
    )


async def async_get_erc20_name_by_block(
    token: spec.ERC20Reference,
    *,
    blocks: typing.Iterable[spec.BlockNumberReference],
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[str]:
    """get name of an erc20 across multiple blocks"""
    return await erc20_generic.async_erc20_eth_call_by_block(
        function_name='name',
        token=token,
        blocks=blocks,
        context=context,
        **rpc_kwargs,
    )


#
# # symbol
#


def _decode_raw_symbol(data: str | None, none_value: str | None = None) -> str:
    """special case decode of ancient non-compliant implementations of symbol"""

    if data is None:
        if none_value is None:
            none_value = ''
        return none_value
    elif len(data) == 66:
        return binary_utils.binary_to_text(data).strip('\x00')
    elif len(data) == 0 or data == '0x':
        return ''
    else:
        as_binary = binary_utils.to_binary(data)
        as_str: str = abi_utils.abi_decode(as_binary, '(string)')[0]
        return as_str


async def async_get_erc20_symbol(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to: str | None = None,
    **rpc_kwargs: typing.Any,
) -> str:
    """get symbol of an erc20"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='erc20_metadata', context=context
    )

    if read_cache:
        from ctc import db

        token = await async_get_erc20_address(token, context=context)
        result: spec.ERC20Metadata | None = await db.async_query_erc20_metadata(
            address=token, context=context
        )
        if result is not None and result['symbol'] is not None:
            return result['symbol']

    symbol_raw = await erc20_generic.async_erc20_eth_call(
        function_name='symbol',
        token=token,
        block=block,
        decode_response=False,
        context=context,
        convert_reverts_to_none=(convert_reverts_to is not None),
        **rpc_kwargs,
    )
    symbol = _decode_raw_symbol(symbol_raw, none_value=convert_reverts_to)

    if write_cache:
        if result is not None:
            result['symbol'] = symbol
        else:
            data = {'address': token, 'symbol': symbol}
            if typing.TYPE_CHECKING:
                result = typing.cast(spec.ERC20Metadata, data)
            else:
                result = data

        await db.async_intake_erc20_metadata(context=context, **result)

    return symbol


async def async_get_erc20s_symbols(
    tokens: typing.Iterable[spec.ERC20Reference],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    convert_reverts_to: str | None = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[str]:
    """get symbol of multiple erc20s"""
    results = await erc20_generic.async_erc20s_eth_calls(
        function_name='symbol',
        tokens=tokens,
        block=block,
        decode_response=False,
        context=context,
        convert_reverts_to_none=(convert_reverts_to is not None),
        **rpc_kwargs,
    )
    return [
        _decode_raw_symbol(result, none_value=convert_reverts_to)
        for result in results
    ]


async def async_get_erc20_symbol_by_block(
    token: spec.ERC20Reference,
    *,
    blocks: typing.Iterable[spec.BlockNumberReference],
    context: spec.Context = None,
    convert_reverts_to: str | None = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[str]:
    """get symbol of an erc20 across multiple blocks"""
    results = await erc20_generic.async_erc20_eth_call_by_block(
        function_name='symbol',
        token=token,
        blocks=blocks,
        decode_response=False,
        context=context,
        convert_reverts_to_none=(convert_reverts_to is not None),
        **rpc_kwargs,
    )
    return [
        _decode_raw_symbol(result, none_value=convert_reverts_to)
        for result in results
    ]




from .erc20_events import *
from .erc20_generic import *
from .erc20_metadata import *
from .erc20_normalize import *
from .erc20_spec import *
from .erc20_state import *
from .erc20_summary import *



from __future__ import annotations

import typing

from ctc import spec
from ... import binary_utils
from .. import abi_coding_utils
from . import event_abi_parsing


@typing.overload
def decode_event_topics(
    topics: typing.Sequence[spec.BinaryData],
    *,
    use_names: typing.Literal[False],
    event_abi: typing.Optional[spec.EventABI] = None,
    indexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    indexed_names: typing.Optional[list[str]] = None,
) -> list[str]:
    ...


@typing.overload
def decode_event_topics(
    topics: typing.Sequence[spec.BinaryData],
    *,
    use_names: typing.Literal[True],
    event_abi: typing.Optional[spec.EventABI] = None,
    indexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    indexed_names: typing.Optional[list[str]] = None,
) -> dict[str, str]:
    ...


@typing.overload
def decode_event_topics(
    topics: typing.Sequence[spec.BinaryData],
    *,
    event_abi: typing.Optional[spec.EventABI] = None,
    indexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    indexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
) -> typing.Union[list[str], dict[str, str]]:
    ...


def decode_event_topics(
    topics: typing.Sequence[spec.BinaryData],
    *,
    event_abi: typing.Optional[spec.EventABI] = None,
    indexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    indexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
) -> typing.Union[list[str], dict[str, str]]:
    """decode topic data (i.e. indexed data) of event"""

    # get abi
    if indexed_types is None:
        if event_abi is None:
            raise Exception('must specify event_abi')
        indexed_types = event_abi_parsing.get_event_indexed_types(event_abi)

    # decode
    decoded_topics: list[str] = []
    for topic, indexed_type in zip(topics[1:], indexed_types):
        # only decode value types
        if (
            indexed_type in ['bytes', 'string']
            or indexed_type.endswith(']')
            or indexed_type.endswith(')')
        ):
            decoded_topics.append(topic)  # type: ignore
        else:
            topic = binary_utils.to_binary(topic)
            decoded_topic = abi_coding_utils.abi_decode(topic, indexed_type)
            decoded_topics.append(decoded_topic)

    # package output
    if not use_names:
        return decoded_topics
    else:
        if indexed_names is None:
            if event_abi is None:
                raise Exception('must specify event_abi')
            indexed_names = event_abi_parsing.get_event_indexed_names(event_abi)
        return dict(zip(indexed_names, decoded_topics))


@typing.overload
def decode_event_unindexed_data(
    data: spec.BinaryData,
    *,
    use_names: typing.Literal[False],
    event_abi: typing.Optional[spec.EventABI] = None,
    unindexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    unindexed_names: typing.Optional[list[str]] = None,
) -> list[str]:
    ...


@typing.overload
def decode_event_unindexed_data(
    data: spec.BinaryData,
    *,
    use_names: typing.Literal[True],
    event_abi: typing.Optional[spec.EventABI] = None,
    unindexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    unindexed_names: typing.Optional[list[str]] = None,
) -> dict[str, str]:
    ...


@typing.overload
def decode_event_unindexed_data(
    data: spec.BinaryData,
    *,
    event_abi: typing.Optional[spec.EventABI] = None,
    unindexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    unindexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
) -> typing.Union[list[str], dict[str, str]]:
    ...


def decode_event_unindexed_data(
    data: spec.BinaryData,
    *,
    event_abi: typing.Optional[spec.EventABI] = None,
    unindexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    unindexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
) -> typing.Union[list[str], dict[str, str]]:
    """decode the unindexed data of event"""

    # gather metadata
    if unindexed_types is None:
        if event_abi is None:
            raise Exception('must specify event_abi')
        unindexed_types = event_abi_parsing.get_event_unindexed_types(event_abi)

    # decode data
    data = binary_utils.to_binary(data)
    decoded = abi_coding_utils.abi_decode(
        data, '(' + ','.join(unindexed_types) + ')'
    )

    # package outputs
    if not use_names:
        return list(decoded)
    else:
        if unindexed_names is None:
            if event_abi is None:
                raise Exception('must specify event_abi')
            unindexed_names = event_abi_parsing.get_event_unindexed_names(
                event_abi
            )
        return dict(zip(unindexed_names, decoded))


def normalize_event(
    event: spec.RawLog,
    event_abi: spec.EventABI,
    *,
    arg_prefix: str | None = 'arg__',
) -> spec.NormalizedLog:
    """normalize raw log data into decoded semantic event data"""

    # decode event args
    decoded_topics = decode_event_topics(
        topics=event['topics'],
        event_abi=event_abi,
        use_names=True,
    )
    decoded_data = decode_event_unindexed_data(
        data=event['data'], event_abi=event_abi, use_names=True
    )

    # remove keys
    remove_keys = ['data', 'topics', 'removed']
    normalized = {k: v for k, v in event.items() if k not in remove_keys}

    # rename keys
    normalized['contract_address'] = normalized['address']

    # add additional keys
    normalized['event_name'] = event_abi['name']
    normalized['event_hash'] = event['topics'][0]

    # add event args
    # args either stored in 'args' key or directly in normalized event
    if arg_prefix is None:
        arg_container: typing.MutableMapping[typing.Any, typing.Any] = {}
        normalized['args'] = arg_container
        arg_prefix = ''
    else:
        arg_container = normalized
    for event_args in [decoded_topics, decoded_data]:
        for arg_name, arg_value in event_args.items():
            key = arg_prefix + arg_name
            if key in arg_container:
                raise Exception('event key collision: ' + str(key))
            arg_container[key] = arg_value

    return normalized




from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ctc import db
from . import chainlink_schema_defs


async def async_upsert_feed(
    feed: typing.Mapping[typing.Any, typing.Any],
    conn: toolsql.AsyncConnection,
    *,
    context: spec.Context = None,
) -> None:

    feed = dict(feed, address=feed['address'].lower())

    table = db.get_table_schema('chainlink_feeds', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=feed,
        upsert=True,
    )


async def async_upsert_feeds(
    feeds: typing.Sequence[typing.Mapping[str, typing.Any]],
    conn: toolsql.AsyncConnection,
    *,
    context: spec.Context = None,
) -> None:

    feeds = [dict(feed, address=feed['address'].lower()) for feed in feeds]

    table = db.get_table_schema('chainlink_feeds', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=feeds,
        upsert=True,
    )


async def async_upsert_aggregator_update(
    *,
    feed: spec.Address,
    aggregator: spec.Address,
    block_number: int,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    row = {
        'feed': feed.lower(),
        'aggregator': aggregator.lower(),
        'block_number': block_number,
    }

    table = db.get_table_schema(
        'chainlink_aggregator_updates', context=context
    )
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=row,
        upsert=True,
    )


async def async_upsert_aggregator_updates(
    *,
    conn: toolsql.AsyncConnection,
    updates: typing.Sequence[chainlink_schema_defs._FeedAggregatorUpdate],
    context: spec.Context = None,
) -> None:

    # convert to lower case
    updates = [
        {
            'feed': update['feed'].lower(),
            'aggregator': update['aggregator'].lower(),
            'block_number': update['block_number'],
        }
        for u, update in list(enumerate(updates))
    ]

    table = db.get_table_schema(
        'chainlink_aggregator_updates', context=context
    )
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=updates,
    )


async def async_select_feed(
    *,
    conn: toolsql.AsyncConnection,
    address: spec.Address | None = None,
    name: str | None = None,
    asset: str | None = None,
    context: spec.Context = None,
) -> chainlink_schema_defs.ChainlinkFeed | None:
    table = db.get_table_schema('chainlink_feeds', context=context)

    where_equals = {
        'address': address,
        'name': name,
        'asset': asset,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }

    return await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals=where_equals,
        output_format='single_dict_or_none',
    )


async def async_select_feeds(
    *,
    conn: toolsql.AsyncConnection,
    address: spec.Address | None = None,
    name: str | None = None,
    asset: str | None = None,
    addresses: typing.Sequence[str] | None = None,
    context: spec.Context = None,
) -> typing.Sequence[chainlink_schema_defs.ChainlinkFeed | None] | None:

    table = db.get_table_schema('chainlink_feeds', context=context)

    where_equals = {
        'address': address,
        'name': name,
        'asset': asset,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }

    if addresses is not None:
        where_in = {'address': addresses}
    else:
        where_in = None

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals=where_equals,
        where_in=where_in,
    )

    return result  # type: ignore


async def async_select_aggregator_updates(
    *,
    conn: toolsql.AsyncConnection,
    feed: spec.Address,
    aggregator: spec.Address | None = None,
    block_number: int | None = None,
    context: spec.Context = None,
) -> typing.Sequence[chainlink_schema_defs._FeedAggregatorUpdate] | None:

    table = db.get_table_schema(
        'chainlink_aggregator_updates', context=context
    )

    if feed is not None:
        feed = feed.lower()
    if aggregator is not None:
        aggregator = aggregator.lower()

    where_equals = {
        'feed': feed,
        'aggregator': aggregator,
        'block_number': block_number,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals=where_equals,
    )

    return result  # type: ignore


async def async_delete_feed(
    conn: toolsql.AsyncConnection,
    *,
    address: spec.Address | None = None,
    name: str | None = None,
    asset: str | None = None,
    context: spec.Context = None,
) -> None:

    where_equals = {
        'address': address,
        'name': name,
        'asset': asset,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }
    if len(where_equals) == 0:
        raise Exception('must specify which feeds to delete')

    table = db.get_table_schema('chainlink_feeds', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals=where_equals,
    )


async def async_delete_aggregator_updates(
    conn: toolsql.AsyncConnection,
    *,
    feed: spec.Address,
    aggregator: spec.Address,
    block_number: int,
    context: spec.Context = None,
) -> None:

    where_equals = {
        'feed': feed,
        'aggregator': aggregator,
        'block_number': block_number,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }
    if len(where_equals) == 0:
        raise Exception('must specify which feeds to delete')

    table = db.get_table_schema(
        'chainlink_aggregator_updates', context=context
    )

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals=where_equals,
    )




from __future__ import annotations

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


def sync_eth_get_transaction_count(
    from_address: spec.Address,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    if block_number is None:
        block_number = 'latest'
    request = rpc_constructors.construct_eth_get_transaction_count(
        from_address=from_address,
        block_number=block_number,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_transaction_count(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_get_transaction_by_hash(
    transaction_hash: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_transaction_by_hash(
        transaction_hash=transaction_hash
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_transaction_by_hash(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


def sync_eth_get_transaction_by_block_hash_and_index(
    block_hash: spec.BinaryData,
    transaction_index: spec.BinaryData,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = (
        rpc_constructors.construct_eth_get_transaction_by_block_hash_and_index(
            block_hash=block_hash,
            transaction_index=transaction_index,
        )
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_transaction_by_block_hash_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


def sync_eth_get_transaction_by_block_number_and_index(
    block_number: spec.BlockNumberReference,
    transaction_index: spec.BinaryData,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_transaction_by_block_number_and_index(
        block_number=block_number,
        transaction_index=transaction_index,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_transaction_by_block_number_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


def sync_eth_get_transaction_receipt(
    transaction_hash: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_transaction_receipt(
        transaction_hash=transaction_hash,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_transaction_receipt(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


def sync_eth_get_block_transaction_count_by_hash(
    block_hash: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = (
        rpc_constructors.construct_eth_get_block_transaction_count_by_hash(
            block_hash=block_hash,
        )
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_block_transaction_count_by_hash(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_get_block_transaction_count_by_number(
    block_number: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = (
        rpc_constructors.construct_eth_get_block_transaction_count_by_number(
            block_number=block_number,
        )
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_block_transaction_count_by_number(
        response=response,
        decode_response=decode_response,
    )




from __future__ import annotations

import typing

import toolcli

from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': keccack_command,
        'help': 'compute keccak hash of data\n\nby default, data treated as hex if it starts with "0x", or treated as text otherwise',
        'args': [
            {'name': 'data', 'help': 'data to hash'},
            {
                'name': '--text',
                'action': 'store_true',
                'help': 'treat input data as text instead of hex',
            },
            {
                'name': '--hex',
                'action': 'store_true',
                'help': 'treat data as hex',
            },
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'omit "0x" prefix on output',
            },
        ],
        'examples': {
            '0x678acb67': 'take hash of hex data',
            'vitalik': 'take hash of text',
        },
    }


def keccack_command(*, data: str, text: bool, hex: bool, raw: bool) -> None:
    if text:
        hex = False
    elif hex:
        pass
    else:
        hex = data.startswith('0x')

    if hex:

        if typing.TYPE_CHECKING:
            casted = typing.cast(spec.GenericBinaryData, data)
        else:
            casted = data

        keccak = evm.keccak(casted, output_format='prefix_hex')
    else:
        keccak = evm.keccak_text(data)

    if raw:
        if not keccak.startswith('0x'):
            raise Exception('wrong format')
        keccak = keccak[2:]

    print(keccak)



from __future__ import annotations

import typing

import ctc
from ctc import spec

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_block_intervals(
    interval_size: tooltime.Timelength,
    *,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    n_intervals: int | None = None,
    window_size: tooltime.Timelength | None = None,
    blocks_at: typing.Literal['start', 'middle', 'end'] = 'middle',
    context: spec.Context | None = None,
    contexts: typing.Sequence[spec.Context] | spec.Context | None = None,
    prefix_network: bool | None = None,
) -> spec.DataFrame:
    """return DataFrame of time intervals with block numbers and timestamps

    TODO: adapt features of async_sample_blocks() below into this function
    """

    import asyncio
    import ctc.config
    import tooltime
    import polars as pl

    # organize contexts
    if contexts is None:
        contexts = [context]
    elif isinstance(contexts, dict):
        contexts = [contexts]
    elif isinstance(contexts, list):
        if len(contexts) == 0:
            contexts = [context]
    else:
        raise Exception("unknown contexts format: " + str(type(contexts)))

    if start_block is not None:
        if len(contexts) != 1:
            raise Exception('invalid context for using start_block')
        start_time = await ctc.async_get_block_timestamp(
            start_block,
            context=contexts[0],
        )
    if end_block is not None:
        if len(contexts) != 1:
            raise Exception('invalid context for using end_block')
        end_time = await ctc.async_get_block_timestamp(
            end_block,
            context=contexts[0],
        )

    df = tooltime.get_interval_df(
        interval_size=interval_size,
        start_time=start_time,
        end_time=end_time,
        n_intervals=n_intervals,
        window_size=window_size,
    )

    # get blocks
    if blocks_at == 'start':
        timestamps = df['start_timestamp']
    elif blocks_at == 'end':
        timestamps = df['end_timestamp']
    elif blocks_at == 'middle':
        timestamps = df['middle_timestamp']
    else:
        raise Exception()

    # get column names
    if prefix_network is None:
        prefix_network = len(contexts) > 1
    if not prefix_network and len(contexts) > 1:
        raise Exception(
            'you must use prefix_network when using multiple contexts'
        )
    if prefix_network:
        names = [
            ctc.config.get_context_network_name(context) + '_block'
            for context in contexts
        ]
    else:
        names = ['block']

    # gather block numbers
    coroutines = [
        ctc.async_get_blocks_of_timestamps(
            timestamps.to_list(), context=context
        )
        for context in contexts
    ]
    contexts_blocks = await asyncio.gather(*coroutines)

    # create dataframe
    block_df = pl.DataFrame(dict(zip(names, contexts_blocks)))
    return df.with_columns(block_df)

# async def async_sample_blocks(
#     *,
#     start_block: spec.BlockNumberReference | None = None,
#     end_block: spec.BlockNumberReference | None = None,
#     include_latest_if_trimmed: bool = True,
#     start_time: tooltime.Timestamp | None = None,
#     end_time: tooltime.Timestamp | None = None,
#     n_samples: int | None = None,
#     sample_interval: tooltime.Timelength | None = None,
#     window_size: tooltime.Timelength | None = None,
#     align_to: typing.Literal['start', 'end'] = 'start',
#     include_misaligned_bound: bool = False,
#     include_misaligned_overflow: bool = False,
#     context: spec.Context = None,
# ) -> typing.Sequence[int]:
#     """create sample of blocks matching input sampling parameters"""

#     import tooltime

#     if start_block is not None:
#         start_time = await block_to_timestamp.async_get_block_timestamp(
#             start_block,
#             context=context,
#         )
#     if end_block is not None:
#         end_time = await block_to_timestamp.async_get_block_timestamp(
#             end_block, context=context
#         )

#     float_timestamps = tooltime.sample_timestamps(
#         start_time=start_time,
#         end_time=end_time,
#         n_samples=n_samples,
#         sample_interval=sample_interval,
#         window_size=window_size,
#         align_to=align_to,
#         include_misaligned_bound=include_misaligned_bound,
#         include_misaligned_overflow=include_misaligned_overflow,
#     )
#     timestamps: typing.Sequence[int] = [
#         round(timestamp) for timestamp in float_timestamps
#     ]

#     # time to latest block timestamp
#     latest_block = await block_crud.async_get_block('latest', context=context)
#     latest_block_timestamp = latest_block['timestamp']
#     n_original_timestamps = len(timestamps)
#     timestamps = [
#         timestamp
#         for timestamp in timestamps
#         if timestamp <= latest_block_timestamp
#     ]
#     trimmed = len(timestamps) != n_original_timestamps

#     # get blocks of timestamps
#     blocks = await timestamp_to_block.async_get_blocks_of_timestamps(
#         timestamps=timestamps, context=context
#     )

#     # include latest if trimmed
#     if trimmed and include_latest_if_trimmed:
#         latest_block_number = latest_block['number']
#         if blocks[-1] != latest_block_number:
#             blocks.append(latest_block_number)

#     return blocks




from __future__ import annotations

import asyncio
import time
import typing

import aiohttp
import toolstr

import ctc.config
from ctc import spec
from . import coingecko_db

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class CoinGeckoRatelimit(TypedDict):
        lock: asyncio.Lock | None
        last_request_time: float
        rps: float
        enabled: bool


_cg_ratelimit: CoinGeckoRatelimit = {
    'lock': None,
    'last_request_time': 0,
    'rps': 50 / 60,
    'enabled': False,
}


urls = {
    'token_list': 'https://api.coingecko.com/api/v3/coins/list',
    'platform_list': 'https://api.coingecko.com/api/v3/coins/list?include_platform=true',
    'token_info': 'https://api.coingecko.com/api/v3/coins/{token_id}',
    'token_market_chart': 'https://api.coingecko.com/api/v3/coins/{token_id}/market_chart?vs_currency=usd&days={days}',
}


async def _async_sleep_for_cg_ratelimit() -> None:

    sleep_time = (
        _cg_ratelimit['last_request_time']
        + 1 / _cg_ratelimit['rps']
        - time.time()
    )
    if sleep_time > 0:
        print('sleeping for', sleep_time, 'seconds')
        await asyncio.sleep(sleep_time)
    _cg_ratelimit['last_request_time'] = time.time()


#
# # lists of tokens
#


async def async_get_token_list(
    *,
    context: spec.Context | None = None,
    update: bool | None = None
) -> typing.Sequence[coingecko_db.CoingeckoToken]:

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='coingecko', context=context
    )

    if read_cache and not update:
        result = await coingecko_db.async_query_tokens(context={})
        if isinstance(result, list) and len(result) > 0:
            return result

    token_list = await _async_get_token_list_from_server(include_platform=False)
    for item in token_list:
        item['market_cap_rank'] = None

    # update db
    if update is None:
        update = True
    if update:
        print('intaking ' + str(len(token_list)) + ' coingecko tokens')
        await coingecko_db.async_intake_tokens(token_list)

    return token_list


async def _async_get_token_list_from_server(
    include_platform: bool = False,
) -> typing.Sequence[typing.Any]:
    if include_platform:
        url = urls['platform_list']
    else:
        url = urls['token_list']

    lock: asyncio.Lock | None = _cg_ratelimit['lock']
    if lock is None:
        lock = asyncio.Lock()
        _cg_ratelimit['lock'] = lock
    async with lock:
        await _async_sleep_for_cg_ratelimit()

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                return await response.json()  # type: ignore


#
# # token symbol --> token id
#


async def async_get_token_name(query: str, *, context: spec.Context = None) -> str:
    """given a token symbol query top token id"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='coingecko', context=context
    )

    if read_cache:
        result = await coingecko_db.async_query_tokens(
            symbol_query=query,
            name_query=query,
            context={},
        )
        if result is not None and len(result) > 0:

            # check for direct symbol matches
            query = query.lower()
            symbol_matches = [
                result for result in result if result['symbol'].lower() == query
            ]
            if len(symbol_matches) >= 1:
                return symbol_matches[0]['name']

            return result[0]['name']

    raise Exception()
    # return await async_get_token_id_from_server(query)


async def async_get_token_id(
    query: str, *, context: spec.Context = None
) -> str:
    """given a token symbol query top token id"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='coingecko', context=context
    )

    if read_cache:
        result = await coingecko_db.async_query_tokens(
            symbol_query=query, name_query=query, context={}
        )
        if result is not None and len(result) > 0:

            # check for direct symbol matches
            query = query.lower()
            symbol_matches = [
                result for result in result if result['symbol'].lower() == query
            ]
            if len(symbol_matches) >= 1:
                return symbol_matches[0]['id']

            return result[0]['id']

    return await async_get_token_id_from_server(query)


async def async_get_token_id_from_server(symbol: str) -> str:
    """return first result from server token list"""

    token_ids = await async_get_token_ids_from_server(symbol)

    if len(token_ids) == 0:
        raise Exception('cannot find token_id')
    elif len(token_ids) > 1:
        # raise Exception('too many token_ids for symbol')
        return token_ids[0]
    else:
        return token_ids[0]


async def async_get_token_ids_from_server(
    symbol: str,
) -> typing.Sequence[str]:
    token_ids_by_symbol = await _async_get_token_ids_by_symbol_from_server()
    token_ids = token_ids_by_symbol.get(symbol.lower(), [])
    return token_ids


async def _async_get_token_ids_by_symbol_from_server() -> typing.Mapping[
    str, typing.Sequence[typing.Any]
]:

    # get token list
    token_list = await _async_get_token_list_from_server()

    # process result
    tokens_by_symbol: typing.MutableMapping[
        str, typing.MutableSequence[typing.Any]
    ] = {}
    for item in token_list:
        tokens_by_symbol.setdefault(item['symbol'], [])
        tokens_by_symbol[item['symbol']].append(item['id'])

    return tokens_by_symbol


#
# # token info
#


async def async_get_token_info(
    *,
    symbol: str | None = None,
    token_id: str | None = None,
) -> typing.Mapping[typing.Any, typing.Any]:
    url_template = urls['token_info']

    if token_id is None:
        if symbol is None:
            raise Exception('must specify token_id or symbol')
        token_id = await async_get_token_id(symbol)

    url = url_template.format(token_id=token_id)

    # lock: asyncio.Lock | None = _cg_ratelimit['lock']
    # if lock is None:
    #     lock = asyncio.Lock()
    #     _cg_ratelimit['lock'] = lock
    # async with lock:
    # await _async_sleep_for_cg_ratelimit()

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.json()  # type: ignore


async def async_get_market_chart(
    *,
    symbol: str | None = None,
    token_id: str | None = None,
    days: str | int | None = None,
) -> typing.Mapping[typing.Any, typing.Any]:

    if days is None:
        days = 'max'

    if token_id is None:
        if symbol is None:
            raise Exception('must specify token_id or symbol')
        token_id = await async_get_token_id(symbol)

    url_template = urls['token_market_chart']
    url = url_template.format(token_id=token_id, days=str(days))

    # lock: asyncio.Lock | None = _cg_ratelimit['lock']
    # if lock is None:
    #     lock = asyncio.Lock()
    #     _cg_ratelimit['lock'] = lock
    # async with lock:
    # await _async_sleep_for_cg_ratelimit()

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            data: typing.Mapping[typing.Any, typing.Any] = await response.json()

    if data.get('status', {}).get('error_code') == 429:
        raise Exception(
            'coingecko ratelimit reached, wait a bit or obtain an API key'
        )

    return data


#
# # data summary
#


async def async_print_token_data_summary(
    *,
    token_id: str | None = None,
    query: str | None = None,
    verbose: bool = False,
    update: bool = False,
    days: int | str | None = None,
) -> None:

    if token_id is None:
        if query is None:
            raise Exception('must specify token_id or query')
        token_id = await async_get_token_id(query)

    token_info_coroutine = asyncio.create_task(
        async_get_token_info(token_id=token_id)
    )
    market_chart_coroutine = asyncio.create_task(
        async_get_market_chart(token_id=token_id, days=days)
    )
    token_info = await token_info_coroutine
    market_chart = await market_chart_coroutine

    market_data = token_info['market_data']
    price = market_data['current_price'].get('usd')
    fdv = market_data['fully_diluted_valuation'].get('usd')
    atl = market_data['atl']['usd']
    ath = market_data['ath']['usd']
    market_cap = market_data['market_cap']['usd']

    dollar_format = {
        'prefix': '$',
        'trailing_zeros': True,
        'decimals': 2,
    }

    rows: typing.MutableSequence[typing.Sequence[typing.Any]] = []
    rows.append(['market cap rank', token_info['market_cap_rank']])
    if price is not None:
        rows.append(['price', toolstr.format(price, **dollar_format)])  # type: ignore
    rows.append(['ATL', toolstr.format(atl, **dollar_format)])  # type: ignore
    rows.append(['ATH', toolstr.format(ath, **dollar_format)])  # type: ignore
    rows.append(['total volume', market_data['total_volume']['usd']])
    rows.append(['market cap', toolstr.format(market_cap, **dollar_format)])  # type: ignore
    if fdv is not None:
        rows.append(['FDV', toolstr.format(fdv, **dollar_format)])  # type: ignore
    rows.append(['circulating supply', market_data['circulating_supply']])
    rows.append(['total supply', market_data['total_supply']])

    from ctc import cli

    styles = cli.get_cli_styles()

    toolstr.print_text_box(
        token_info['name'] + ' Coingecko data', style=styles['title']
    )
    print()
    toolstr.print_header('Statistics', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        indent=4,
        border=styles['comment'],
        column_styles=[styles['option'], styles['description']],
    )

    platforms = token_info['platforms']
    if (
        verbose
        and len(platforms) > 0
        and not (len(platforms) == 1 and tuple(platforms) == ('',))
    ):
        print()
        print()
        toolstr.print_header('Platforms', style=styles['title'])
        rows = []
        for platform, address in token_info['platforms'].items():
            row: typing.Sequence[typing.Any] = (platform, address)
            rows.append(row)
        toolstr.print_table(
            rows,
            indent=4,
            border=styles['comment'],
            column_styles=[styles['option'], styles['description']],
        )

    print()
    toolstr.print_header('Charts', style=styles['title'])
    prices_times = [datum[0] / 1000 for datum in market_chart['prices']]
    total_volumes_times = [
        datum[0] / 1000 for datum in market_chart['total_volumes']
    ]
    market_caps_times = [
        datum[0] / 1000 for datum in market_chart['market_caps']
    ]
    prices = [datum[1] for datum in market_chart['prices']]
    total_volumes = [datum[1] for datum in market_chart['total_volumes']]
    market_caps = [datum[1] for datum in market_chart['market_caps']]

    plots: typing.Sequence[
        tuple[str, typing.Sequence[float | int], typing.Sequence[float | int]]
    ] = [
        ('Price', prices_times, prices),
        ('Market Cap', market_caps_times, market_caps),
        ('Volume', total_volumes_times, total_volumes),
    ]
    for title, xvals, yvals in plots:
        plot = toolstr.render_line_plot(
            xvals=xvals,
            yvals=yvals,
            n_rows=10,
            n_columns=60,
            line_style=styles['description'],
            chrome_style=styles['comment'],
            tick_label_style=styles['metavar'],
            yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
            char_dict=ctc.config.get_cli_chart_charset(),
        )
        print()

        if yvals[0] != 0:
            delta = yvals[-1] / yvals[0] - 1
        else:
            delta = float('inf')
        formatted_delta = toolstr.format(delta, percentage=True, signed=True)
        formatted_delta = '(' + formatted_delta + ')'
        formatted_delta = toolstr.add_style(formatted_delta, styles['comment'])
        toolstr.print(
            toolstr.hjustify(title + ' ' + formatted_delta, 'center', 84)[:80],
            indent=4,
            style=styles['title'],
        )
        toolstr.print(plot, indent=4)
        print()

        if title != plots[-1][0]:
            print()

    if verbose:
        print()
        print()
        toolstr.print_header('Tickers', style=styles['title'])
        rows = []
        for ticker in token_info['tickers'][:20]:
            if 'target_coin_id' in ticker:
                target = ticker['target_coin_id']
            else:
                target = ticker['target']
            row = [
                ticker['market']['name'],
                target,
                toolstr.format(ticker['volume'], **dollar_format),  # type: ignore
            ]
            rows.append(row)
        if len(token_info['tickers']) > 20:
            rows.append(['...', '...', '...'])
        labels = [
            'platform',
            'token',
            'volume',
        ]
        toolstr.print_table(
            rows,
            labels=labels,
            indent=4,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={'volume': styles['description']},
        )

    updated_time_str = 'metadata updated: ' + token_info['last_updated']
    updated_time_str.replace('T', ' ')
    timespan_str = 'timespan = ' + str(days) + ' days'
    timespan_str = toolstr.hjustify(timespan_str, 'right', 73)
    combined_str = updated_time_str + timespan_str[len(updated_time_str) :]
    toolstr.print(combined_str, style=styles['comment'])


async def async_print_coin_quotient_summary(
    coin1: str,
    coin2: str,
    *,
    days: int | None = None,
) -> None:
    import numpy as np
    from ctc import cli

    chart1 = await async_get_market_chart(symbol=coin1, days=days)
    chart2 = await async_get_market_chart(symbol=coin2, days=days)

    coin1_times, coin1_prices = np.array(chart1['prices']).T
    coin2_times, coin2_prices = np.array(chart2['prices']).T

    coin1_times /= 1000
    coin2_times /= 1000

    coin1_over_coin2_times, coin1_over_coin2_prices = _compute_token_quotient(
        coin1_times=coin1_times,
        coin2_times=coin2_times,
        coin1_prices=coin1_prices,
        coin2_prices=coin2_prices,
    )
    coin2_over_coin1_times, coin2_over_coin1_prices = _compute_token_quotient(
        coin1_times=coin2_times,
        coin2_times=coin1_times,
        coin1_prices=coin2_prices,
        coin2_prices=coin1_prices,
    )

    styles = cli.get_cli_styles()

    toolstr.print(
        toolstr.hjustify(coin1 + ' / ' + coin2, 'center', 70),
        indent=4,
        style=styles['title'],
    )
    plot = toolstr.render_line_plot(
        xvals=coin1_over_coin2_times,  # type: ignore
        yvals=coin1_over_coin2_prices,  # type: ignore
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        char_dict=ctc.config.get_cli_chart_charset(),
    )
    toolstr.print(plot, indent=4)

    print()
    print()
    print()
    toolstr.print(
        toolstr.hjustify(coin2 + ' / ' + coin1, 'center', 70),
        indent=4,
        style=styles['title'],
    )
    plot = toolstr.render_line_plot(
        xvals=coin2_over_coin1_times,  # type: ignore
        yvals=coin2_over_coin1_prices,  # type: ignore
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        char_dict=ctc.config.get_cli_chart_charset(),
    )
    toolstr.print(plot, indent=4)


def _compute_token_quotient(
    *,
    coin1_times: spec.NumpyArray,
    coin2_times: spec.NumpyArray,
    coin1_prices: spec.NumpyArray,
    coin2_prices: spec.NumpyArray,
) -> tuple[spec.NumpyArray, spec.NumpyArray]:
    """returns coin1_prices / coin_prices using simple backlook interpolation"""

    import numpy as np

    indices = np.searchsorted(coin1_times, coin2_times)
    nonzero_mask = (indices != 0) * (indices < len(coin1_times))
    nonzero_indices = indices[nonzero_mask]
    comparison_prices = coin1_prices[nonzero_indices]
    coin1_over_coin2_prices = comparison_prices / coin2_prices[nonzero_mask]
    return coin2_times[nonzero_mask], coin1_over_coin2_prices




from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc.cli.cli_utils import cli_alias_utils


def add_cli_aliases(
    *,
    styles: toolcli.StyleTheme,
    headless: bool,
    skip_aliases: bool,
) -> None:

    print()
    print()
    toolstr.print('## Installing CLI Aliases', style=styles['title'])

    alias_status = cli_alias_utils.get_paths_alias_status()
    current = all(status == 'current' for status in alias_status.values())

    if current:
        print()
        cli_alias_utils.print_alias_status(include_title=False)
    else:
        print()
        cli_alias_utils.print_paths_alias_status()
        print()
        print(
            'ctc can install cli aliases to make many commands quicker to type'
        )
        print()
        toolstr.print(
            'For example, you can type '
            + toolstr.add_style('4byte 0xa9059cbb', styles['option'])
            + ' instead of '
            + toolstr.add_style('ctc 4byte 0xa9059cbb', styles['option'])
        )
        print()
        print('ctc has aliases prepared for the following commands:')
        print()
        cli_alias_utils.print_aliases()
        print()

        if skip_aliases:
            default = 'no'
        else:
            default = 'yes'

        if toolcli.input_yes_or_no(
            'Do you want to install these aliases? ',
            style=styles['metavar'],
            default=default,
            headless=headless,
        ):
            cli_alias_utils.install_aliases(confirm=True, headless=headless)
            print()
            print('Aliases installed')



from __future__ import annotations

import typing

from ctc import config
from ctc import spec


def get_dex_names_of_factories(
    *,
    network: spec.NetworkReference | None = None,
    context: spec.Context = None,
) -> typing.Mapping[spec.Address, str]:
    """get mapping of factory_address -> dex_name"""

    if network is None:
        network = config.get_context_chain_id(context)

    if network in (1, 'ethereum'):
        return {
            '0x5c69bee701ef814a2b6a3edd4b1652cb9cc5aa6f': 'Uniswap V2',
            '0x1f98431c8ad98523631ae4a59f267346ea31f984': 'Uniswap V3',
            '0xba12222222228d8ba445958a75a0704d566bf2c8': 'Balancer',
            '0xb9fc157394af804a3578134a6585c0dc9cc990d4': 'Curve',
            '0x0959158b6040d32d04c301a72cbfd6b39e21c9ae': 'Curve',
            '0xf18056bbd320e96a48e3fbf8bc061322531aac99': 'Curve',
            '0x90e00ace148ca3b23ac1bc8c240c2a7dd9c2d7f5': 'Curve',
            '0x8f942c20d02befc377d41445793068908e2250d0': 'Curve',
            '0xbabe61887f1de2713c6f97e567623453d3c79f67': 'Curve',
            '0xc0aee478e3658e2610c5f7a4a2e1777ce9e4f2ac': 'Sushi',
        }
    else:
        raise Exception(
            'dex pool factory map not available for network: ' + str(network)
        )


def get_dex_name_of_factory(
    factory: spec.Address,
    *,
    context: spec.Context = None,
) -> str:
    """get dex_name of factory address"""

    names_of_factories = get_dex_names_of_factories(context=context)
    return names_of_factories[factory]




from __future__ import annotations

from ctc import spec


def convert_rpc_block_to_db_block(
    rpc_block: spec.RPCBlock | spec.DBBlock,
) -> spec.DBBlock:
    """convert a raw block from an rpc request into a standard ctc block"""
    return {
        'number': rpc_block['number'],
        'hash': rpc_block['hash'],
        'timestamp': rpc_block['timestamp'],
        'miner': rpc_block['miner'],
        'extra_data': rpc_block['extra_data'],
        'base_fee_per_gas': rpc_block.get('base_fee_per_gas'),
        'gas_limit': rpc_block['gas_limit'],
        'gas_used': rpc_block['gas_used'],
    }

